"""Classes for spectral data"""
# pylint: disable=no-member

from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
import copy
from functools import partial
import math
from numbers import Integral, Real
from typing import (
    Any,
    Literal,
    Optional,
    TypeVar,
    overload,
)
import warnings

import numpy as np
from pint import DimensionalityError, Quantity
from scipy.ndimage import correlate1d, gaussian_filter

from euphonic.broadening import (
    FWHM_TO_SIGMA,
    ErrorFit,
    KernelShape,
    variable_width_broadening,
)
from euphonic.io import (
    _obj_from_json_file,
    _obj_to_dict,
    _obj_to_json_file,
    _process_dict,
)
from euphonic.readers.castep import read_phonon_dos_data
from euphonic.ureg import ureg
from euphonic.util import dedent_and_fill, zips
from euphonic.validate import _check_constructor_inputs, _check_unit_conversion

CallableQuantity = Callable[[Quantity], Quantity]
XTickLabels = list[tuple[int, str]]

OneSpectrumMetadata = dict[str, str | int]


class WidthTypeError(TypeError): ...


class Spectrum(ABC):
    """Base class for a spectral data: do not use directly"""
    T = TypeVar('T', bound='Spectrum')

    def __setattr__(self, name: str, value: Any) -> None:
        _check_unit_conversion(self, name, value,
                               ['x_data_unit', 'y_data_unit'])
        super().__setattr__(name, value)

    def _set_data(self, data: Quantity, attr_name: str) -> None:
        setattr(self, f'_{attr_name}_data',
                np.array(data.magnitude, dtype=np.float64))
        setattr(self, f'_internal_{attr_name}_data_unit', str(data.units))
        setattr(self, f'{attr_name}_data_unit', str(data.units))

    @property
    def x_data(self) -> Quantity:
        """x-axis data with units"""
        return ureg.Quantity(self._x_data, self._internal_x_data_unit,
                             ).to(self.x_data_unit, 'reciprocal_spectroscopy')

    @x_data.setter
    def x_data(self, value: Quantity) -> None:
        self.x_data_unit = str(value.units)
        self._x_data = value.to(self._internal_x_data_unit).magnitude

    @property
    def y_data(self) -> Quantity:
        """y-axis data with units"""
        return ureg.Quantity(self._y_data, self._internal_y_data_unit).to(
            self.y_data_unit, 'reciprocal_spectroscopy')

    @y_data.setter
    def y_data(self, value: Quantity) -> None:
        self.y_data_unit = str(value.units)
        self._y_data = value.to(self._internal_y_data_unit).magnitude

    def __imul__(self: T, other: Real) -> T:
        """Scale spectral data in-place"""
        self._y_data *= other
        return self

    def __mul__(self: T, other: Real) -> T:
        """Get a new spectrum with scaled data"""
        new_spec = self.copy()
        new_spec *= other
        return new_spec

    @abstractmethod
    def copy(self: T) -> T:
        """Get an independent copy of spectrum"""

    @property
    def x_tick_labels(self) -> XTickLabels:
        """x-axis tick labels (e.g. high-symmetry point locations)"""
        return self._x_tick_labels

    @x_tick_labels.setter
    def x_tick_labels(self, value: XTickLabels) -> None:
        err_msg = (
            'x_tick_labels should be of type Sequence[Tuple[int, str]] e.g. '
            '[(0, "label1"), (5, "label2")]'
        )
        if value is not None:
            if isinstance(value, Sequence):
                for elem in value:
                    if not (isinstance(elem, tuple)
                            and len(elem) == 2
                            and isinstance(elem[0], Integral)
                            and isinstance(elem[1], str)):
                        raise TypeError(err_msg)
                # Ensure indices in x_tick_labels are plain ints as
                # np.int64/32 etc. are not JSON serializable
                value = [(int(idx), label) for idx, label in value]
            else:
                raise TypeError(err_msg)
        self._x_tick_labels = value

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Write to dict using euphonic.io._obj_to_dict"""

    @classmethod
    @abstractmethod
    def from_dict(cls: type[T], d: dict[str, Any]) -> T:
        """Initialise a Spectrum object from dictionary"""

    def to_json_file(self, filename: str) -> None:
        """
        Write to a JSON file. JSON fields are equivalent to
        from_dict keys

        Parameters
        ----------
        filename : str
            Name of the JSON file to write to
        """
        _obj_to_json_file(self, filename)

    @classmethod
    def from_json_file(cls: type[T], filename: str) -> T:
        """
        Read from a JSON file. See from_dict for required fields

        Parameters
        ----------
        filename : str
            The file to read from
        """
        type_dict = {'x_tick_labels': tuple}
        return _obj_from_json_file(cls, filename, type_dict)

    @abstractmethod
    def _split_by_indices(self: T, indices: Sequence[int] | np.ndarray,
                          ) -> list[T]:
        """Split data along x axis at given indices"""

    def _split_by_tol(self: T, btol: float = 10.0) -> list[T]:
        """Split data along x-axis at detected breakpoints"""
        diff = np.diff(self.x_data)
        median = np.median(diff)
        breakpoints = np.where((diff / median) > btol)[0] + 1
        return self._split_by_indices(breakpoints)

    @staticmethod
    def _ranges_from_indices(indices: Sequence[int] | np.ndarray,
                             ) -> list[tuple[int, Optional[int]]]:
        """Convert a series of breakpoints to a series of slice ranges"""
        if len(indices) == 0:
            ranges = [(0, None)]
        else:
            ranges = [(0, indices[0])]
            if len(indices) > 1:
                ranges = ranges + [(indices[i], indices[i+1])
                                   for i in range(len(indices) - 1)]
            ranges = [*ranges, (indices[-1], None)]
        return ranges

    @staticmethod
    def _cut_x_ticks(x_tick_labels: XTickLabels | None,
                     x0: int,
                     x1: int | None) -> XTickLabels | None:
        """Crop x labels to new x range, shifting indices accordingly"""
        if x_tick_labels is None:
            return None

        if x1 is None:
            x1 = float('inf')

        return [(int(x - x0), label)
                for (x, label) in x_tick_labels if x0 <= x < x1]

    def split(self: T, indices: Sequence[int] | np.ndarray = None,
              btol: float | None = None) -> list[T]:
        """Split to multiple spectra

        Data may be split by index. Alternatively, x-axis data may be
        split automatically by searching for unusually large gaps in
        energy values.  These would usually correspond to disconnected
        special-points in a band-structure diagram.

        Parameters
        ----------
        indices
            positions in data of breakpoints
        btol
            parameter used to identify breakpoints. This is a ratio
            between the gap in values and the median gap between
            neighbouring x-values. If neither indices nor btol is
            specified, this is set to 10.0.

        Returns
        -------
        split_spectra
            Separated spectrum regions. If passed to the appropriate
            functions in euphonic.plot this would be interpreted as a
            series of subplots.
        """

        if indices is None:
            if btol is None:
                btol = 10.0
            return self._split_by_tol(btol=btol)

        if btol is not None:
            msg = 'Cannot set both indices and btol'
            raise ValueError(msg)
        return self._split_by_indices(indices)

    @classmethod
    def _broaden_data(cls,
                      data: np.ndarray,
                      bin_centres: Sequence[np.ndarray],
                      widths: Sequence[float|None],
                      shape: KernelShape = 'gauss',
                      *,
                      method: Optional[Literal['convolve']] = None,
                      width_convention: Literal['fwhm', 'std'] = 'fwhm',
                      ) -> np.ndarray:
        """
        Broaden data along each axis by widths, returning the broadened
        data. Data can be 1D or 2D, and  the length of bin_centres and
        widths should match the number of dimensions in data.
        bin_centres and widths are assumed to be in the same units
        """
        def _has_bad_type(width: Any) -> bool:
            return width is not None and not isinstance(width, Real)

        if any(map(_has_bad_type, widths)):
            msg = ('Inappropriate type found, widths for _broaden_data '
                   'must be Real (e.g. float) or None. Instead we have: ['
                   + ', '.join(type(width).__name__ for width in widths)
                   + '].')
            raise WidthTypeError(msg)

        shape_opts = ('gauss', 'lorentz')
        if shape not in shape_opts:
            msg = (
                f'Invalid value for shape, got {shape}, '
                f'should be one of {shape_opts}'
            )
            raise ValueError(msg)
        method_opts = ('convolve', None)
        if method not in method_opts:
            msg = (
                f'Invalid value for method, got {method}, '
                f'should be one of {method_opts}'
            )
            raise ValueError(msg)

        # We only want to check for unequal bins if using a method that
        # is not correct for unequal bins (currently this is the only
        # option but leave the 'if' anyway for future implementations)
        axes = ['x_data', 'y_data']
        unequal_bin_axes = []
        for ax, (width, bin_data) in enumerate(zips(widths, bin_centres)):
            if width is not None:
                bin_widths = np.diff(bin_data)
                if not np.all(np.isclose(bin_widths, bin_widths[0])):
                    unequal_bin_axes += [axes[ax]]
        if len(unequal_bin_axes) > 0:
            msg = dedent_and_fill(f"""
            {" and ".join(unequal_bin_axes)} bin widths are not equal, so
            broadening by convolution may give incorrect results.
            """)
            if method is None:
                raise ValueError(
                    msg + ' If you still want to broaden by convolution '
                    'please explicitly use the method="convolve" option.')

            warnings.warn(msg, stacklevel=3)

        if shape == 'gauss':
            width_to_bin = partial(cls._gaussian_width_to_bin_sigma,
                                   width_convention=width_convention)
            sigmas = list(map(width_to_bin, widths, bin_centres))

            data_broadened = gaussian_filter(data, sigmas, mode='constant')

        elif shape == 'lorentz':
            if width_convention != 'fwhm':
                msg = 'Lorentzian function width must be specified as FWHM'
                raise ValueError(msg)
            data_broadened = data
            for ax, (width, bin_data) in enumerate(zips(widths, bin_centres)):
                if width is not None:
                    broadening = _distribution_1d(bin_data, width)
                    data_broadened = correlate1d(data_broadened, broadening,
                                                 mode='constant',
                                                 axis=ax)
        return data_broadened

    @staticmethod
    def _gaussian_width_to_bin_sigma(
        width: float | None,
        ax_bin_centres: np.ndarray,
        width_convention: Literal['fwhm', 'std'],
    ) -> float:
        """
        Convert a Gaussian FWHM to sigma in units of the mean ax bin size

        Parameters
        ----------
        width
            The Gaussian broadening width parameter FWHM or sigma (STD)
        ax_bin_centres
            Shape (n_bins,) float np.ndarray.
            The bin centres along the axis the broadening is applied to
        width_convention
            Specify width as full-width-half-maximum 'fwhm' or standard
            deviation 'std'

        Returns
        -------
        sigma
            Sigma in units of mean ax bin size
        """
        if width is None:
            return 0.

        match width_convention:
            case 'fwhm':
                sigma = width * FWHM_TO_SIGMA
            case 'std':
                sigma = width
            case _:
                msg = "Width convention must be 'std' or 'fwhm'"
                raise ValueError(msg)

        mean_bin_size = np.mean(np.diff(ax_bin_centres))
        return sigma / mean_bin_size

    @staticmethod
    def _bin_edges_to_centres(bin_edges: Quantity) -> Quantity:
        return bin_edges[:-1] + 0.5*np.diff(bin_edges)

    @staticmethod
    def _bin_centres_to_edges(
            bin_centres: Quantity,
            restrict_range: bool = True,
    ) -> Quantity:
        if restrict_range:
            return np.concatenate((
                [bin_centres[0]],
                (bin_centres[1:] + bin_centres[:-1])/2,
                [bin_centres[-1]],
            ))

        half_diff = np.diff(bin_centres) * 0.5
        return np.concatenate(
            (
                [bin_centres[0] - half_diff[0]],
                bin_centres[:-1] + half_diff,
                [bin_centres[-1] + half_diff[-1]],
            ))

    @staticmethod
    def _is_bin_edge(data_length: int, bin_length: int) -> bool:
        """Determine if axis data are bin edges or centres"""
        if bin_length == data_length + 1:
            return True
        if bin_length == data_length:
            return False
        msg = (
            f'Unexpected data axis length {data_length} '
            f'for bin axis length {bin_length}'
        )
        raise ValueError(msg)

    def get_bin_edges(self, *, restrict_range: bool = True) -> Quantity:
        """
        Get x-axis bin edges. If the size of x_data is one element larger
        than y_data, x_data is assumed to contain bin edges, but if x_data
        is the same size, x_data is assumed to contain bin centres and
        a conversion is made.

        In this case, bin edges are assumed to be halfway between bin centres.

        Parameters
        ----------
        restrict_range
            If True (default), the bin edges will not go outside the existing
            data bounds so the first and last bins may be half-size. This may
            be desirable for plotting.  Otherwise, the outer bin edges will
            extend from the initial data range.
        """
        # Need to use -1 index for y_data so it also works for
        # Spectrum1DCollection which has y_data shape (n_spectra, bins)
        if self._is_bin_edge(self.y_data.shape[-1], self.x_data.shape[0]):
            return self.x_data
        return self._bin_centres_to_edges(
            self.x_data, restrict_range=restrict_range)

    def get_bin_centres(self) -> Quantity:
        """
        Get x-axis bin centres. If the size of x_data is the same size
        as y_data, x_data is assumed to contain bin centres, but if x_data
        is one element larger, x_data is assumed to contain bin edges and
        a conversion is made. In this conversion, the bin centres are
        assumed to be in the middle of each bin, which may not be an
        accurate assumption in the case of differently sized bins.
        """
        # Need to use -1 index for y_data so it also works for
        # Spectrum1DCollection which has y_data shape (n_spectra, bins)
        if self._is_bin_edge(self.y_data.shape[-1], self.x_data.shape[0]):
            return self._bin_edges_to_centres(self.x_data)
        return self.x_data

    def get_bin_widths(self, *, restrict_range: bool = True) -> Quantity:
        """
        Get x-axis bin widths

        Parameters
        ----------
        restrict_range
            If True, bin edges will be clamped to the input data range; if
            False, they may be extrapolated beyond the initial range of bin
            centres.

            False is usually preferable, this default behaviour is for backward
            compatibility and will be removed in a future version.

        """
        return np.diff(self.get_bin_edges(restrict_range=restrict_range))

    def assert_regular_bins(self,
                            message: str = '',
                            rtol: float = 1e-5,
                            atol: float = 0.,
                            restrict_range: bool = True,
                            ) -> None:
        """Raise AssertionError if x-axis bins are not evenly spaced.

        Note that the positional arguments are different from
        Spectrum2D.assert_regular_bins: it is strongly recommended to only use
        keyword arguments with this method.

        Parameters
        ----------
        message
            Text appended to ValueError for more informative output.

        rtol
            Relative tolerance for 'close enough' values

        atol
            Absolute tolerance for 'close enough' values. Note this is a bare
            float and follows the stored units of the bins.

        restrict_range
            If True, bin edges will be clamped to the input data range; if
            False, they may be extrapolated beyond the initial range of bin
            centres.

            You should use the value which is consistent with calls to
            get_bin_widths() or get_bin_edges().

        """
        bin_widths = self.get_bin_widths(restrict_range=restrict_range)
        # Need to cast to magnitude to use isclose() with atol before Pint 0.21
        if not np.all(np.isclose(bin_widths.magnitude, bin_widths.magnitude[0],
                                 rtol=rtol, atol=atol)):
            raise AssertionError('Not all x-axis bins are the same width. '
                                 + message)


class Spectrum1D(Spectrum):
    """
    For storing generic 1D spectra e.g. density of states

    Attributes
    ----------
    x_data
        Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The x_data
        points (if size == (n_x_data,)) or x_data bin edges (if size
        == (n_x_data + 1,))
    y_data
        Shape (n_x_data,) float Quantity. The plot data in y
    x_tick_labels
        Sequence[Tuple[int, str]] or None. Special tick labels e.g. for
        high-symmetry points. The int refers to the index in x_data the
        label should be applied to
    metadata
        Dict[str, Union[int, str]]. Contains metadata about the
        spectrum. Keys should be strings and values should be strings
        or integers
        There are some functional keys:

          - 'label' : str. This is used label lines on a 1D plot
    """
    T = TypeVar('T', bound='Spectrum1D')

    def __init__(self, x_data: Quantity, y_data: Quantity,
                 x_tick_labels: Optional[XTickLabels] = None,
                 metadata: Optional[dict[str, int | str]] = None,
                 ) -> None:
        """
        Parameters
        ----------
        x_data
            Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The
            x_data points (if size == (n_x_data,)) or x_data bin edges
            (if size == (n_x_data + 1,))
        y_data
            Shape (n_x_data,) float Quantity. The plot data in y
        x_tick_labels
            Special tick labels e.g. for high-symmetry points. The int
            refers to the index in x_data the label should be applied to
        metadata
            Contains metadata about the spectrum. Keys should be
            strings and values should be strings or integers
            There are some functional keys:

              - 'label' : str. This is used label lines on a 1D plot
        """
        _check_constructor_inputs(
            [y_data, x_tick_labels, metadata],
            [Quantity, [list, type(None)], [dict, type(None)]],
            [(-1,), (), ()],
            ['y_data', 'x_tick_labels', 'metadata'])
        ny = len(y_data)
        _check_constructor_inputs(
            [x_data], [Quantity],
            [[(ny,), (ny+1,)]], ['x_data'])
        self._set_data(x_data, 'x')
        self._set_data(y_data, 'y')
        self.x_tick_labels = x_tick_labels
        self.metadata = {} if metadata is None else metadata

    def __add__(self, other: 'Spectrum1D') -> 'Spectrum1D':
        """
        Sums the y_data of two Spectrum1D objects together,
        their x_data axes must be equal, and their y_data must
        have compatible units and the same number of y_data
        entries

        Any metadata key/value pairs that are common to both
        spectra are retained, any others are discarded
        """
        # pylint: disable=import-outside-toplevel
        from .collections import Spectrum1DCollection
        spec_col = Spectrum1DCollection.from_spectra([self, other])
        return spec_col.sum()

    def _split_by_indices(self: T,
                          indices: Sequence[int] | np.ndarray,
                          ) -> list[T]:
        """Split data along x-axis at given indices"""
        ranges = self._ranges_from_indices(indices)

        return [type(self)(self.x_data[x0:x1], self.y_data[x0:x1],
                           x_tick_labels=self._cut_x_ticks(self.x_tick_labels,
                                                           x0, x1),
                           metadata=self.metadata)
                for x0, x1 in ranges]

    def copy(self: T) -> T:
        """Get an independent copy of spectrum"""
        return type(self)(np.copy(self.x_data),
                          np.copy(self.y_data),
                          x_tick_labels=copy.copy(self.x_tick_labels),
                          metadata=copy.deepcopy(self.metadata))

    def to_dict(self) -> dict[str, Any]:
        """
        Convert to a dictionary. See Spectrum1D.from_dict for details on
        keys/values

        Returns
        -------
        dict
        """
        return _obj_to_dict(self, ['x_data', 'y_data', 'x_tick_labels',
                                   'metadata'])

    def to_text_file(self, filename: str,
                     fmt: Optional[str | Sequence[str]] = None) -> None:
        """
        Write to a text file. The header contains metadata and unit
        information, the first column contains x_data and the second
        column contains y_data. Note that text files written in this
        format cannot be read back in by Euphonic.

        Parameters
        ----------
        filename
            Name of the text file to write to
        fmt
            A format specifier or sequence of specifiers (one for each
            column), to be passed to numpy.savetxt
        """
        # pylint: disable=import-outside-toplevel
        from .collections import Spectrum1DCollection
        spec = Spectrum1DCollection.from_spectra([self])
        spec.to_text_file(filename, fmt)

    @classmethod
    def from_dict(cls: type[T], d: dict[str, Any]) -> T:
        """
        Convert a dictionary to a Spectrum1D object

        Parameters
        ----------
        d
            A dictionary with the following keys/values:

            - 'x_data': (n_x_data,) or (n_x_data + 1,) float ndarray
            - 'x_data_unit': str
            - 'y_data': (n_x_data,) float ndarray
            - 'y_data_unit': str

            There are also the following optional keys:

            - 'x_tick_labels': list of (int, string) tuples
            - 'metadata': dict

        Returns
        -------
        spectrum
        """
        d = _process_dict(d, quantities=['x_data', 'y_data'],
                          optional=['x_tick_labels', 'metadata'])
        return cls(d['x_data'], d['y_data'], x_tick_labels=d['x_tick_labels'],
                   metadata=d['metadata'])

    @classmethod
    def from_castep_phonon_dos(cls: type[T], filename: str,
                               element: Optional[str] = None) -> T:
        """
        Reads DOS from a CASTEP .phonon_dos file

        Parameters
        ----------
        filename
            The path and name of the .phonon_dos file to read
        element
            Which element's PDOS to read. If not supplied reads
            the total DOS.
        """
        data = read_phonon_dos_data(filename)
        metadata = {}
        if element is None:
            element = 'Total'
        else:
            metadata['species'] = element
        metadata['label'] = element

        return cls(ureg.Quantity(data['dos_bins'],
                                 units=data['dos_bins_unit']),
                   ureg.Quantity(data['dos'][element], units=data['dos_unit']),
                   metadata=metadata)

    @overload
    def broaden(self: T, x_width: Quantity,
                shape: KernelShape = 'gauss',
                method: Optional[Literal['convolve']] = None,
                width_convention: Literal['fwhm', 'std'] = 'fwhm',
                ) -> T: ...

    @overload
    def broaden(self: T, x_width: CallableQuantity,
                shape: KernelShape = 'gauss',
                method: Optional[Literal['convolve']] = None,
                width_lower_limit: Optional[Quantity] = None,
                width_convention: Literal['fwhm', 'std'] = 'fwhm',
                width_interpolation_error: float = 0.01,
                width_fit: ErrorFit = 'cheby-log',
                ) -> T: ...

    def broaden(self: T, x_width,
                shape='gauss',
                method=None,
                width_lower_limit=None,
                width_convention='fwhm',
                width_interpolation_error=0.01,
                width_fit='cheby-log',
                ) -> T:
        """
        Broaden y_data and return a new broadened spectrum object

        Parameters
        ----------
        x_width
            Scalar float Quantity, giving broadening FWHM for all y data, or
            a function handle accepting Quantity consistent with x-axis as
            input and returning FWHM corresponding to input array. This would
            typically be an energy-dependent resolution function.
        shape
            One of {'gauss', 'lorentz'}. The broadening shape
        method
            Can be None or 'convolve'. Currently the only broadening
            method available is convolution with a broadening kernel
            but this may not produce correct results for unequal bin
            widths. To use convolution anyway, explicitly set
            method='convolve'
        width_lower_limit
            Set a lower bound to width obtained calling x_width function. By
            default, this is equal to bin width. To disable, set to -Inf.
        width_convention
            By default ('fwhm'), x_width is interpreted as full-width
            half-maximum. Set to 'std' to instead define standard deviation.
        width_interpolation_error
            When x_width is a callable function, variable-width broadening is
            implemented by an approximate kernel-interpolation scheme. This
            parameter determines the target error of the kernel approximations.
        width_fit
            Select parametrisation of kernel width spacing to
            width_interpolation_error.  'cheby-log' is recommended: for shape
            'gauss', 'cubic' is also available.

        Returns
        -------
        broadened_spectrum
            A new Spectrum1D object with broadened y_data

        Raises
        ------
        ValueError
            If shape is not one of the allowed strings
        ValueError
            If method is None and bins are not of equal size
        """
        if isinstance(x_width, Quantity):
            y_broadened = self._broaden_data(
                self.y_data.magnitude,
                [self.get_bin_centres().magnitude],
                [x_width.to(self.x_data_unit).magnitude],
                shape=shape, method=method, width_convention=width_convention)
            y_broadened = ureg.Quantity(y_broadened, units=self.y_data_unit)

        elif isinstance(x_width, Callable):
            self.assert_regular_bins(
                message=(
                    'Broadening with convolution requires a '
                    'regular sampling grid.'
                ),
                restrict_range=False,
            )
            y_broadened = variable_width_broadening(
                self.get_bin_edges(restrict_range=False),
                self.get_bin_centres(),
                x_width,
                (self.y_data * self.get_bin_widths(restrict_range=False)[0]),
                width_lower_limit=width_lower_limit,
                width_convention=width_convention,
                adaptive_error=width_interpolation_error,
                shape=shape,
                fit=width_fit,
            )
        else:
            msg = 'x_width must be a Quantity or Callable'
            raise TypeError(msg)

        new_spectrum = self.copy()
        new_spectrum.y_data = y_broadened
        return new_spectrum


class Spectrum2D(Spectrum):
    """
    For storing generic 2D spectra e.g. S(Q,w)

    Attributes
    ----------
    x_data
        Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The x_data
        points (if size == (n_x_data,)) or x_data bin edges (if size
        == (n_x_data + 1,))
    y_data
        Shape (n_y_data,) or (n_y_data + 1,) float Quantity. The y_data
        bin points (if size == (n_y_data,)) or y_data bin edges (if size
        == (n_y_data + 1,))
    z_data
        Shape (n_x_data, n_y_data) float Quantity. The plot data in z
    x_tick_labels
        Sequence[Tuple[int, str]] or None. Special tick labels e.g. for
        high-symmetry points. The int refers to the index in x_data the
        label should be applied to
    metadata
        Dict[str, Union[int, str]]. Contains metadata about the
        spectrum. Keys should be strings and values should be strings
        or integers
    """
    T = TypeVar('T', bound='Spectrum2D')

    def __init__(self, x_data: Quantity, y_data: Quantity,
                 z_data: Quantity,
                 x_tick_labels: Optional[XTickLabels] = None,
                 metadata: Optional[OneSpectrumMetadata] = None,
                 ) -> None:
        """
        Parameters
        ----------
        x_data
            Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The
            x_data points (if size == (n_x_data,)) or x_data bin edges
            (if size == (n_x_data + 1,))
        y_data
            Shape (n_y_data,) or (n_y_data + 1,) float Quantity. The
            y_data bin points (if size == (n_y_data,)) or y_data bin
            edges (if size == (n_y_data + 1,))
        z_data
            Shape (n_x_data, n_y_data) float Quantity. The plot data in z
        x_tick_labels
            Special tick labels e.g. for high-symmetry points. The int
            refers to the index in x_data the label should be applied to
        metadata
            Contains metadata about the spectrum. Keys should be
            strings and values should be strings or integers.
        """
        _check_constructor_inputs(
            [z_data, x_tick_labels, metadata],
            [Quantity, [list, type(None)], [dict, type(None)]],
            [(-1, -1), (), ()],
            ['z_data', 'x_tick_labels', 'metadata'])
        nx = z_data.shape[0]
        ny = z_data.shape[1]
        _check_constructor_inputs(
            [x_data, y_data],
            [Quantity, Quantity],
            [[(nx,), (nx + 1,)], [(ny,), (ny + 1,)]],
            ['x_data', 'y_data'])
        self._set_data(x_data, 'x')
        self._set_data(y_data, 'y')
        self.x_tick_labels = x_tick_labels
        self._set_data(z_data, 'z')
        self.metadata = {} if metadata is None else metadata

    @property
    def z_data(self) -> Quantity:
        """z-axis data with units"""
        return ureg.Quantity(
            self._z_data, self._internal_z_data_unit,
        ).to(self.z_data_unit, 'reciprocal_spectroscopy')

    @z_data.setter
    def z_data(self, value: Quantity) -> None:
        self.z_data_unit = str(value.units)
        self._z_data = value.to(self._internal_z_data_unit).magnitude

    def __imul__(self: T, other: Real) -> T:
        """Scale spectral data in-place"""
        self.z_data = self.z_data * other
        return self

    def __setattr__(self, name: str, value: Any) -> None:
        _check_unit_conversion(self, name, value,
                               ['z_data_unit'])
        super().__setattr__(name, value)

    def _split_by_indices(self,
                          indices: Sequence[int] | np.ndarray,
                          ) -> list[T]:
        """Split data along x-axis at given indices"""
        ranges = self._ranges_from_indices(indices)
        return [type(self)(self.x_data[x0:x1], self.y_data,
                           self.z_data[x0:x1, :],
                           x_tick_labels=self._cut_x_ticks(self.x_tick_labels,
                                                           x0, x1),
                           metadata=self.metadata)
                for x0, x1 in ranges]

    def broaden(self: T,
                x_width: Optional[Quantity | CallableQuantity] = None,
                y_width: Optional[Quantity | CallableQuantity] = None,
                shape: KernelShape = 'gauss',
                method: Optional[Literal['convolve']] = None,
                x_width_lower_limit: Quantity = None,
                y_width_lower_limit: Quantity = None,
                width_convention: Literal['fwhm', 'std'] = 'fwhm',
                width_interpolation_error: float = 0.01,
                width_fit: ErrorFit = 'cheby-log',
                ) -> T:
        """
        Broaden z_data and return a new broadened Spectrum2D object

        Callable functions can be used to access variable-width broadening.
        In this case, broadening is implemented with a kernel-interpolating
        approximate scheme.

        Parameters
        ----------
        x_width
            Either a scalar float Quantity representing the broadening width,
            or a callable function that accepts and returns Quantity consistent
            with x-axis units.
        y_width
            Either a scalar float Quantity representing the broadening width,
            or a callable function that accepts and returns Quantity consistent
            with y-axis units.
        shape
            One of {'gauss', 'lorentz'}. The broadening shape
        method
            Can be None or 'convolve'. Currently the only broadening
            method available is convolution with a broadening kernel,
            but this may not produce correct results for unequal bin
            widths. To use convolution anyway, explicitly set
            method='convolve'
        x_width_lower_limit
            Set a lower bound to width obtained calling x_width function. By
            default, this is equal to x bin width. To disable, set to -Inf.
        y_width_lower_limit
            Set a lower bound to width obtained calling y_width function. By
            default, this is equal to y bin width. To disable, set to -Inf.
        width_convention
            By default ('fwhm'), widths are interpreted as full-width
            half-maximum. Set to 'std' to instead define standard deviation.
        width_interpolation_error
            When x_width is a callable function, variable-width broadening is
            implemented by an approximate kernel-interpolation scheme. This
            parameter determines the target error of the kernel approximations.
        width_fit
            Select parametrisation of kernel width spacing to
            width_interpolation_error. 'cheby-log' is recommended: for shape
            'gauss', 'cubic' is also available.

        Returns
        -------
        broadened_spectrum
            A new Spectrum2D object with broadened z_data

        Raises
        ------
        ValueError
            If shape is not one of the allowed strings
        ValueError
            If method is None and bins are not of equal size
        """
        # First apply fixed-width broadening; this can be applied to both axes
        # in one function call if both are specified.
        widths_in_bin_units = [None]*2

        if isinstance(x_width, Quantity):
            try:
                self.assert_regular_bins(bin_ax='x', message=(
                    'Broadening by convolution may give incorrect results.'))
            except AssertionError as e:
                warnings.warn(str(e), UserWarning, stacklevel=1)
            widths_in_bin_units[0] = x_width.to(self.x_data_unit).magnitude

        if isinstance(y_width, Quantity):
            widths_in_bin_units[1] = y_width.to(self.y_data_unit).magnitude

        if any(widths_in_bin_units):
            bin_centres = [self.get_bin_centres(ax).magnitude
                           for ax in ['x', 'y']]

            z_broadened = self._broaden_data(
                self.z_data.magnitude,
                bin_centres,
                widths_in_bin_units,
                shape=shape,
                method=method,
                width_convention=width_convention)

            spectrum = Spectrum2D(
                np.copy(self.x_data),
                np.copy(self.y_data),
                ureg.Quantity(z_broadened, units=self.z_data_unit),
                copy.copy(self.x_tick_labels),
                copy.deepcopy(self.metadata),
            )

        else:
            spectrum = self

        # Then apply variable-width broadening to any axes with Callable
        for axis, width, lower_limit in (('x', x_width, x_width_lower_limit),
                                         ('y', y_width, y_width_lower_limit)):
            if isinstance(width, Callable):
                spectrum = self._broaden_spectrum2d_with_function(
                    spectrum, width, axis=axis, width_lower_limit=lower_limit,
                    width_convention=width_convention,
                    width_interpolation_error=width_interpolation_error,
                    shape=shape,
                    width_fit=width_fit)

        return spectrum

    @staticmethod
    def _broaden_spectrum2d_with_function(
            spectrum: 'Spectrum2D',
            width_function: Callable[[Quantity], Quantity],
            axis: Literal['x', 'y'] = 'y',
            width_lower_limit: Quantity = None,
            width_convention: Literal['fwhm', 'std'] = 'fwhm',
            width_interpolation_error: float = 1e-2,
            shape: KernelShape = 'gauss',
            width_fit: ErrorFit = 'cheby-log',
    ) -> 'Spectrum2D':
        """
        Apply value-dependent Gaussian broadening to one axis of Spectrum2D
        """
        assert axis in ('x', 'y')

        bins = spectrum.get_bin_edges(bin_ax=axis, restrict_range=False)
        bin_widths = spectrum.get_bin_widths(bin_ax=axis, restrict_range=False)

        if not np.all(np.isclose(bin_widths, bin_widths[0])):
            bin_width = bin_widths.mean()
        else:
            bin_width = bin_widths[0]

        # Input data: rescale to sparse-like data values
        z_data = spectrum.z_data * bin_width

        if axis == 'x':
            z_data = z_data.T

        # Output data: matches input units
        z_broadened = np.empty_like(z_data) * spectrum.z_data.units

        for i, row in enumerate(z_data):
            z_broadened[i] = variable_width_broadening(
                bins,
                spectrum.get_bin_centres(bin_ax=axis),
                width_function,
                row,
                width_lower_limit=width_lower_limit,
                width_convention=width_convention,
                adaptive_error=width_interpolation_error,
                shape=shape,
                fit=width_fit)

        if axis == 'x':
            z_broadened = z_broadened.T

        return Spectrum2D(np.copy(spectrum.x_data),
                          np.copy(spectrum.y_data),
                          z_broadened,
                          copy.copy(spectrum.x_tick_labels),
                          copy.copy(spectrum.metadata))

    def copy(self: T) -> T:
        """Get an independent copy of spectrum"""
        return type(self)(np.copy(self.x_data),
                          np.copy(self.y_data),
                          np.copy(self.z_data),
                          copy.copy(self.x_tick_labels),
                          copy.deepcopy(self.metadata))

    def get_bin_edges(
            self,
            bin_ax: Literal['x', 'y'] = 'x',
            *,
            restrict_range: bool = True,
    ) -> Quantity:
        """
        Get bin edges for the axis specified by bin_ax. If the size of bin_ax
        is one element larger than z_data along that axis, bin_ax is assumed to
        contain bin edges. If they are the same size, bin_ax is assumed to
        contain bin centres and a conversion is made.

        In this case, bin edges are assumed to be halfway between bin centres.

        Parameters
        ----------
        bin_ax
            The axis to get the bin edges for, 'x' or 'y'

        restrict_range
            If True (default), the bin edges will not go outside the existing
            data bounds so the first and last bins may be half-size. This may
            be desirable for plotting.  Otherwise, the outer bin edges will
            extend from the initial data range.
        """
        enum = {'x': 0, 'y': 1}
        bin_data = getattr(self, f'{bin_ax}_data')
        data_ax_len = self.z_data.shape[enum[bin_ax]]
        if self._is_bin_edge(data_ax_len, bin_data.shape[0]):
            return bin_data
        return self._bin_centres_to_edges(
            bin_data, restrict_range=restrict_range)

    def get_bin_centres(self, bin_ax: Literal['x', 'y'] = 'x') -> Quantity:
        """
        Get bin centres for the axis specified by bin_ax. If the size of
        bin_ax is the same size as z_data along that axis, bin_ax is
        assumed to contain bin centres, but if bin_ax is one element
        larger, bin_ax is assumed to contain bin centres and a conversion
        is made. In this conversion, the bin centres are assumed to be in
        the middle of each bin, which may not be an accurate assumption in
        the case of differently sized bins.

        Parameters
        ----------
        bin_ax
            The axis to get the bin centres for, 'x' or 'y'
        """
        enum = {'x': 0, 'y': 1}
        bin_data = getattr(self, f'{bin_ax}_data')
        data_ax_len = self.z_data.shape[enum[bin_ax]]
        if self._is_bin_edge(data_ax_len, bin_data.shape[0]):
            return self._bin_edges_to_centres(bin_data)
        return bin_data

    def get_bin_widths(
            self,
            bin_ax: Literal['x', 'y'] = 'x',
            *,
            restrict_range: bool = True,
    ) -> Quantity:
        """
        Get x-bin widths along specified axis

        Parameters
        ----------
        bin_ax
            Axis of interest, 'x' or 'y'

        restrict_range
            If True, bin edges will be clamped to the input data range; if
            False, they may be extrapolated beyond the initial range of bin
            centres.

            False is usually preferable, this default behaviour is for backward
            compatibility and will be removed in a future version.
        """
        bins = self.get_bin_edges(bin_ax, restrict_range=restrict_range)
        return np.diff(bins)

    def assert_regular_bins(  # pylint: disable=arguments-renamed
            self,
            bin_ax: Literal['x', 'y'],
            message: str = '',
            rtol: float = 1e-5,
            atol: float = 0.,
            restrict_range: bool = True,
    ) -> None:
        """Raise AssertionError if x-axis bins are not evenly spaced.

        Note that the positional arguments are different from
        Spectrum1D.assert_regular_bins: it is strongly recommended to only use
        keyword arguments with this method.

        Parameters
        ----------
        bin_ax
            Axis of interest, 'x' or 'y'

        message
            Text appended to ValueError for more informative output.

        rtol
            Relative tolerance for 'close enough' values

        atol
            Absolute tolerance for 'close enough' values. Note this is a bare
            float and follows the stored units of the bins.

        restrict_range
            If True, bin edges will be clamped to the input data range; if
            False, they may be extrapolated beyond the initial range of bin
            centres.

            You should use the value which is consistent with calls to
            get_bin_widths() or get_bin_edges().
        """
        bin_widths = self.get_bin_widths(bin_ax, restrict_range=restrict_range)
        if not np.all(np.isclose(bin_widths, bin_widths[0],
                                 rtol=rtol, atol=atol)):
            raise AssertionError(
                f'Not all {bin_ax}-axis bins are the same width. ' + message)

    def to_dict(self) -> dict[str, Any]:
        """
        Convert to a dictionary. See Spectrum2D.from_dict for details on
        keys/values

        Returns
        -------
        dict
        """
        return _obj_to_dict(self, ['x_data', 'y_data', 'z_data',
                                   'x_tick_labels', 'metadata'])

    @classmethod
    def from_dict(cls: type[T], d: dict[str, Any]) -> T:
        """
        Convert a dictionary to a Spectrum2D object

        Parameters
        ----------
        d : dict
            A dictionary with the following keys/values:

            - 'x_data': (n_x_data,) or (n_x_data + 1,) float ndarray
            - 'x_data_unit': str
            - 'y_data': (n_y_data,) or (n_y_data + 1,) float ndarray
            - 'y_data_unit': str
            - 'z_data': (n_x_data, n_y_data) float Quantity
            - 'z_data_unit': str

            There are also the following optional keys:

            - 'x_tick_labels': list of (int, string) tuples
            - 'metadata': dict

        Returns
        -------
        spectrum
        """
        d = _process_dict(d, quantities=['x_data', 'y_data', 'z_data'],
                          optional=['x_tick_labels', 'metadata'])
        return cls(d['x_data'], d['y_data'], d['z_data'],
                   x_tick_labels=d['x_tick_labels'],
                   metadata=d['metadata'])


def apply_kinematic_constraints(spectrum: Spectrum2D,
                                e_i: Quantity = None,
                                e_f: Quantity = None,
                                angle_range: tuple[float] = (0, 180.),
                                ) -> Spectrum2D:
    """
    Set events to NaN which violate energy/momentum limits:

      - Energy transfer greater than e_i
      - q outside region accessible for given e_i and angle range

    This requires x_data to be in wavevector units and y_data to be energy.

    Either e_i or e_f should be set, according to direct/instrument
    geometry. The other values will be inferred, interpreting y_data as
    energy transfer.

    Parameters
    ----------
    spectrum
        input 2-D spectrum, with |q| x_data and energy y_data
    e_i
        incident energy of direct-geometry spectrometer
    e_f
        final energy of indirect-geometry spectrometer
    angle_range
        min and max scattering angles (2θ) of detector bank in degrees

    Returns
    -------
    Masked spectrum with inaccessible bins set to NaN in z_data.
    """
    try:
        (1 * spectrum.x_data.units).to('1/angstrom')
    except DimensionalityError as error:
        msg = 'x_data needs to have wavevector units (i.e. 1/length)'
        raise ValueError(msg) from error
    try:
        (1 * spectrum.y_data.units).to('eV', 'spectroscopy')
    except DimensionalityError as error:
        msg = 'y_data needs to have energy (or wavenumber) units'
        raise ValueError(msg) from error

    momentum2_to_energy = 0.5 * (ureg('hbar^2 / neutron_mass')
                                 .to('meV angstrom^2'))

    if (e_i is None) == (e_f is None):
        msg = dedent_and_fill("""
            Exactly one of e_i and e_f should be set.
            (The other value will be derived from energy transfer).
            """)
        raise ValueError(msg)

    if e_i is None:
        # Indirect geometry: final energy is fixed, incident energy unlimited
        e_f = e_f.to('meV')
        e_i = spectrum.get_bin_centres(bin_ax='y').to('meV') + e_f
    elif e_f is None:
        # Direct geometry: incident energy is fixed, max energy transfer = e_i
        e_i = e_i.to('meV')
        e_f = e_i - spectrum.get_bin_centres(bin_ax='y').to('meV')

    k2_i = e_i / momentum2_to_energy
    k2_f = e_f / momentum2_to_energy

    cos_values = np.asarray(
        _get_cos_range(np.asarray(angle_range) * np.pi / 180.))

    # Momentum goes negative where final energy greater than incident
    # energy; detect this as complex component and set extreme q-bounds to
    # enforce conservation of energy

    # (Complex number sqrt separated from units sqrt for compatibility with
    # older library versions; in newer versions this is not necessary.)
    q_bounds = np.sqrt(k2_i + k2_f
                       - 2 * cos_values[:, np.newaxis]
                           * np.sqrt(k2_i.magnitude * k2_f.magnitude,
                                     dtype=complex)
                           * (k2_i.units * k2_f.units)**0.5,
                       )
    q_bounds.magnitude.T[np.any(q_bounds.imag, axis=0)] = [float('Inf'),
                                                           float('-Inf')]
    q_bounds = q_bounds.real

    mask = np.logical_or((spectrum.get_bin_edges(bin_ax='x')[1:, np.newaxis]
                          < q_bounds[0][np.newaxis, :]),
                         (spectrum.get_bin_edges(bin_ax='x')[:-1, np.newaxis]
                          > q_bounds[-1][np.newaxis, :]))

    new_spectrum = spectrum.copy()
    new_spectrum._z_data[mask] = float('nan')

    return new_spectrum


def _get_cos_range(angle_range: tuple[float]) -> tuple[float]:
    """
    Get max and min of cosine function over angle range

    These will either be the cosines of the input angles, or, in the case that
    a cosine max/min point lies within the angle range, 1/-1 respectively.

    Method: the angle range is translated such that it starts within 0-2π; then
    we check for the presence of turning points at π and 2π.
    """
    limiting_values = np.cos(angle_range).tolist()

    shift, lower_angle = divmod(min(angle_range), np.pi * 2)
    upper_angle = max(angle_range) - (shift * 2 * np.pi)
    if lower_angle < np.pi < upper_angle:
        limiting_values.append(-1.)
    if lower_angle < 2 * np.pi < upper_angle:
        limiting_values.append(1.)
    return max(limiting_values), min(limiting_values)


def _distribution_1d(xbins: np.ndarray,
                     xwidth: float,
                     shape: Literal['lorentz'] = 'lorentz',
                     ) -> np.ndarray:
    x = _get_dist_bins(xbins)
    if shape != 'lorentz':
        msg = "Expected shape: 'lorentz'"
        raise ValueError(msg)

    dist = _lorentzian(x, xwidth)
    return dist / np.sum(dist)  # Naively normalise


def _get_dist_bins(bins: np.ndarray) -> np.ndarray:
    # Get bins with same spacing as original, centered on zero
    # and spanning at least twice as far; this ensures a point in bin can be
    # distributed all the way across bin range.
    bin_range = bins[-1] - bins[0]
    nbins = len(bins) * 2 + 1

    return np.linspace(-bin_range, bin_range, nbins, endpoint=True)


def _lorentzian(x: np.ndarray, gamma: float) -> np.ndarray:
    return gamma/(2*math.pi*(np.square(x) + (gamma/2)**2))



#! /usr/bin/env python

"""Bump the current version number in CHANGELOG.rst

Unreleased changes will be moved into a section for the new version number
"""
from argparse import ArgumentParser
from dataclasses import dataclass
from pathlib import Path
import re

from packaging.version import Version
from toolz.itertoolz import partition

REPOSITORY_ADDRESS = 'https://github.com/pace-neutrons/Euphonic'
LINK_STRING = '`{tag} <' + REPOSITORY_ADDRESS + '/compare/{previous_tag}...{compare_tag}>`_'


@dataclass
class Block:
    """CHANGELOG data corresponding to a single release"""
    tag: str
    previous_tag: str
    content: str

    @staticmethod
    def tag_to_header(tag: str, previous_tag: str) -> str:
        """Produce header including github diff link"""
        compare_tag = 'HEAD' if tag == 'Unreleased' else tag
        header = LINK_STRING.format(tag=tag,
                                    previous_tag=previous_tag,
                                    compare_tag=compare_tag)
        return header + '\n' + len(header) * '-'

    def __str__(self) -> str:
        txt = self.tag_to_header(self.tag, self.previous_tag)
        if self.content:
            txt += f'\n\n{self.content}'

        return txt


def parse_changelog(changelog_file: Path) -> list[Block]:
    """Read all sections from changelog file

    Each section should begin with two lines in format

      `TAG <repository_address/compare/PREV_TAG...TAG>`_
      --------------------------------------------------

    except for 'Unreleased' which compares with HEAD
    """

    with changelog_file.open(encoding='utf8') as fd:
        split_text = re.split(
            LINK_STRING.replace('.', r'\.').format(
                tag=r'(\S+)', previous_tag=r'(\S+)', compare_tag=r'\S+',
            )
            + r'\n-+',
            fd.read(),
        )

    # First item is always empty?
    split_text = split_text[1:]

    # From Python 3.12 can replace partition with itertools.batched
    blocks = [Block(tag, prev_tag, content) for tag, prev_tag, content in partition(3, split_text)]

    for block in blocks:
        block.content = block.content.strip()

    return blocks


def bump_version(blocks: list[Block], tag: str) -> None:
    """Create or update a new or existing block with content from "Unreleased" """

    all_tags = [block.tag for block in blocks]

    if all_tags[0] != 'Unreleased':
        msg = 'CHANGELOG should always begin with Unreleased'
        raise ValueError(msg)

    if all_tags[1] != tag:
        # This version is not in the CHANGELOG yet
        previous_tag = all_tags[1]
        blocks.insert(1, Block(tag, previous_tag, ''))

    blocks[1].content = '\n\n'.join(
        txt for txt in (blocks[0].content, blocks[1].content) if txt
    )

    # Update "Unreleased" section
    blocks[0].content = ''
    blocks[0].previous_tag = tag


def to_text(blocks: list[Block]) -> str:
    """Dump blocks to single multiline string"""
    return '\n\n'.join(map(str, blocks))


def get_parser() -> ArgumentParser:
    """Use argparse to get user input"""
    parser = ArgumentParser()
    parser.add_argument('filename', type=Path, help='Input CHANGELOG.rst file')
    parser.add_argument('tag', type=str, help='Tag for new/updated version')
    parser.add_argument('--replace', action='store_true', help='Write to existing file')

    return parser



def main() -> None:
    """Entrypoint"""
    args = get_parser().parse_args()

    if Version(args.tag).is_prerelease:
        print('New version is a pre-release, leave CHANGELOG alone.')
        return

    blocks = parse_changelog(args.filename)
    bump_version(blocks, args.tag)

    if args.replace:
        with open(args.filename, 'w', encoding='utf8') as fd:
            print(to_text(blocks), file=fd)

    else:
        print(to_text(blocks))


if __name__ == '__main__':
    main()



from argparse import ArgumentParser
from typing import Optional

import matplotlib.style
from numpy import sqrt
from numpy.polynomial import Polynomial

from euphonic import ForceConstants, QpointPhononModes, ureg
from euphonic.plot import plot_1d
from euphonic.styles import base_style
from euphonic.util import dedent_and_fill, mode_gradients_to_widths, mp_grid

from .utils import (
    _arrange_pdos_groups,
    _calc_modes_kwargs,
    _compose_style,
    _get_cli_parser,
    _get_energy_bins,
    _get_pdos_weighting,
    _grid_spec_from_args,
    _plot_label_kwargs,
    get_args,
    load_data_from_file,
    matplotlib_save_or_show,
)


def main(params: Optional[list[str]] = None) -> None:
    parser = get_parser()
    args = get_args(parser, params)

    frequencies_only = (args.weighting == 'dos' and args.pdos is None)
    data = load_data_from_file(args.filename, verbose=True,
                               frequencies_only=frequencies_only)

    if not frequencies_only and not isinstance(
            data, (QpointPhononModes, ForceConstants)):
        msg = (
            'Eigenvectors are required to use "--pdos" or '
            'any "--weighting" option other than plain DOS'
        )
        raise TypeError(msg)
    if args.adaptive and not isinstance(data, ForceConstants):
        msg = 'Force constants are required to use --adaptive option'
        raise TypeError(msg)

    if (args.energy_broadening
            and args.adaptive
            and len(args.energy_broadening) == 1):
        if args.adaptive_scale is not None:
            msg = dedent_and_fill("""
                Adaptive scale factor was specified twice; use either
                --adaptive-scale or --energy-broadening.  To add a fixed width
                to adaptive broadening, use --instrument-broadening.""")
            raise ValueError(msg)
        args.adaptive_scale = args.energy_broadening[0]

    elif args.energy_broadening:
        if args.inst_broadening is not None:
            msg = (
                'Broadening width was specified twice; use either'
                '--instrument-broadening or --energy-broadening.'
            )
            raise ValueError(msg)
        args.inst_broadening = args.energy_broadening

    if args.inst_broadening:
        energy_broadening_poly = Polynomial(args.inst_broadening)

    mode_widths = None
    if isinstance(data, ForceConstants):

        recip_length_unit = ureg(f'1 / {args.length_unit}')
        grid_spec = _grid_spec_from_args(data.crystal, grid=args.grid,
                                         grid_spacing=(args.grid_spacing
                                                       * recip_length_unit))
        print('Calculating phonon modes '
              'on {} q-point grid...'.format(
                  ' x '.join([str(x) for x in grid_spec])))
        if args.adaptive:
            cmkwargs = _calc_modes_kwargs(args)
            cmkwargs['return_mode_gradients'] = True
            modes, mode_grads = data.calculate_qpoint_phonon_modes(
                mp_grid(grid_spec), **cmkwargs)
            mode_widths = mode_gradients_to_widths(mode_grads,
                                                   modes.crystal.cell_vectors)
            if args.adaptive_scale:
                mode_widths *= args.adaptive_scale
            if args.inst_broadening and args.shape == 'gauss':
                # Combine instrumental broadening and adaptive sample
                # broadening: the convolution of a Gaussian with a Gaussian is
                # a Gaussian with sigma = sqrt(sigma1^2 + sigma2^2)
                mode_widths = sqrt(
                    mode_widths**2
                    + (energy_broadening_poly(modes.frequencies
                                              .to(args.energy_unit).magnitude,
                                              ) * ureg(args.energy_unit))**2)
        else:
            modes = data.calculate_qpoint_phonon_modes(
                mp_grid(grid_spec), **_calc_modes_kwargs(args))

    else:
        modes = data
    modes.frequencies_unit = args.energy_unit
    ebins = _get_energy_bins(
        modes, args.ebins + 1, emin=args.e_min, emax=args.e_max)

    kwargs = {'mode_widths': mode_widths,
              'adaptive_method': args.adaptive_method,
              'adaptive_error': args.adaptive_error,
              'adaptive_error_fit': args.adaptive_fit}

    if args.weighting == 'dos' and args.pdos is None:
        dos = modes.calculate_dos(ebins, **kwargs)
    else:
        kwargs['weighting'] = _get_pdos_weighting(args.weighting)
        pdos = modes.calculate_pdos(ebins, **kwargs)
        dos = _arrange_pdos_groups(pdos, args.pdos)

    if args.inst_broadening and args.shape == 'gauss' and args.adaptive:
        pass  # Gaussian broadening included with adaptive sampling

    elif (args.inst_broadening and len(energy_broadening_poly) > 1):
        # Variable-width Gaussian broadening
        def energy_broadening_func(x):
            return energy_broadening_poly(x.to(args.energy_unit).magnitude,
                                          ) * ureg(args.energy_unit)

        dos = dos.broaden(x_width=energy_broadening_func,
                          shape=args.shape,
                          method='convolve',
                          width_interpolation_error=args.adaptive_error,
                          width_fit='cheby-log')

    elif args.inst_broadening:
        # Fixed-width broadening
        dos = dos.broaden(energy_broadening_poly.coef[0] * ebins.units,
                          shape=args.shape)

    plot_label_kwargs = _plot_label_kwargs(
        args, default_xlabel=f'Energy / {dos.x_data.units:~P}')

    if args.scale is not None:
        dos *= args.scale

    if args.save_json:
        dos.to_json_file(args.save_json)
    style = _compose_style(user_args=args, base=[base_style])
    with matplotlib.style.context(style):
        _ = plot_1d(dos, ymin=0, **plot_label_kwargs)
        matplotlib_save_or_show(save_filename=args.save_to)


def get_parser() -> ArgumentParser:
    parser, _ = _get_cli_parser(features={'read-fc', 'read-modes', 'mp-grid',
                                          'plotting', 'ebins',
                                          'adaptive-broadening',
                                          'pdos-weighting',
                                          'scaling'})
    parser.description = (
        'Plots a DOS from the file provided. If a force '
        'constants file is provided, a DOS is generated on the Monkhorst-Pack '
        'grid specified by the grid (or grid-spacing) argument.')

    return parser



from .crystal import Crystal
from .debye_waller import DebyeWaller
from .force_constants import ForceConstants
from .qpoint_frequencies import QpointFrequencies
from .qpoint_phonon_modes import QpointPhononModes
from .spectra import Spectrum1D, Spectrum1DCollection, Spectrum2D
from .structure_factor import StructureFactor
from .ureg import Quantity, ureg
from .version import __version__  # noqa: F401

__all__ = [
    'Crystal',
    'DebyeWaller',
    'ForceConstants',
    'QpointFrequencies',
    'QpointPhononModes',
    'Quantity',
    'Spectrum1D',
    'Spectrum1DCollection',
    'Spectrum2D',
    'StructureFactor',
    'ureg',
]



from collections import defaultdict
from collections.abc import Iterator
import re
import struct
from typing import (
    Any,
    BinaryIO,
    NamedTuple,
    Optional,
    TextIO,
)

import numpy as np
from packaging.version import Version

from euphonic.ureg import ureg
from euphonic.util import dedent_and_fill


def read_phonon_dos_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        frequencies_unit: str = 'meV',
        mode_gradients_unit: str = 'meV*angstrom') -> dict[str, Any]:
    """
    Reads data from a .phonon_dos file and returns it in a dictionary

    Parameters
    ----------
    filename
        The path and name of the .phonon file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    frequencies_unit
        The unit to return the frequencies in

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'qpts', 'weights', 'frequencies',
        'frequencies_unit', 'mode_gradients', 'mode_gradients_unit',
        'dos_bins', 'dos_bins_unit', 'dos', 'pdos'
    """
    with open(filename) as f:

        f.readline()  # Skip BEGIN header
        n_atoms = int(f.readline().split()[-1])
        _ = int(f.readline().split()[-1])  # n_species is not used
        n_branches = int(f.readline().split()[-1])
        n_bins = int(f.readline().split()[-1])
        (cell_vectors, atom_r, atom_type, atom_mass) = _read_crystal_info(
            f, n_atoms)
        f.readline()  # Skip BEGIN GRADIENTS

        qpts = np.empty((0, 3))
        weights = np.empty((0,))
        freqs = np.empty((0, n_branches))
        mode_grads = np.empty((0, n_branches))
        for frequency_block in _read_frequency_blocks(
                f,
                n_branches,
                extra_columns=[0],
                terminator=' END GRADIENTS\n'):

            qmode_grad = frequency_block.extra

            qpts = np.concatenate((qpts, [frequency_block.qpt]))
            weights = np.concatenate((weights, [frequency_block.weight]))
            freqs = np.concatenate((freqs, [frequency_block.frequencies]))
            mode_grads = np.concatenate((mode_grads, [qmode_grad.squeeze()]))
        f.readline()  # Skip line

        dos_data = np.loadtxt(f, max_rows=n_bins)

    data_dict: dict[str, Any] = {}
    cry_dict = data_dict['crystal'] = {}
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = (cell_vectors*ureg('angstrom').to(
        cell_vectors_unit)).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    cry_dict['atom_r'] = atom_r
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*(ureg('amu')).to(
        atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit

    data_dict['qpts'] = qpts
    data_dict['weights'] = weights
    data_dict['frequencies'] = ((freqs*(1/ureg.cm)).to(
        frequencies_unit)).magnitude
    data_dict['frequencies_unit'] = frequencies_unit
    # Pint conversion 'angstrom * (1/cm)' to 'angstrom * (meV)' doesn't
    # work so convert to meV first then go to specified units
    mode_grads = (mode_grads * ureg('angstrom')
                  ).to(cell_vectors_unit).magnitude
    data_dict['mode_gradients'] = (
        ((mode_grads * ureg('1/cm')).to('meV') * ureg('angstrom'))
        .to(mode_gradients_unit)
        .magnitude
    )
    data_dict['mode_gradients_unit'] = mode_gradients_unit
    data_dict['dos_bins'] = dos_data[:, 0]*ureg('1/cm').to(
            frequencies_unit).magnitude
    data_dict['dos_bins_unit'] = frequencies_unit

    data_dict['dos'] = {}
    data_dict['dos_unit'] = f'1/{frequencies_unit}'
    dos_conv = ureg('1 / (1/cm)',
                    ).to(data_dict['dos_unit'], 'reciprocal_spectroscopy',
                         ).magnitude

    dos_dict = data_dict['dos']
    dos_dict['Total'] = dos_data[:, 1] * dos_conv

    _, idx = np.unique(atom_type, return_index=True)
    unique_types = atom_type[np.sort(idx)]
    for i, species in enumerate(unique_types):
        dos_dict[str(species)] = dos_data[:, i + 2] * dos_conv

    return data_dict


def read_phonon_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        frequencies_unit: str = 'meV',
        read_eigenvectors: bool = True,
        average_repeat_points: bool = True,
        prefer_non_loto: bool = False,
    ) -> dict[str, Any]:
    """
    Reads data from a .phonon file and returns it in a dictionary

    Parameters
    ----------
    filename
        The path and name of the .phonon file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    frequencies_unit
        The unit to return the frequencies in
    read_eigenvectors
        Whether to read the eigenvectors and return them in the
        dictionary
    average_repeat_points
        If multiple frequency/eigenvectors blocks are provided with the same
        q-point index (i.e. for Gamma-point with LO-TO splitting), scale the
        weights such that these sum to the given weight
    prefer_non_loto
        If multiple frequency/eigenvector blocks are provided with the same
        q-point index and all-but-one include a direction vector, use the data
        from the point without a direction vector. (i.e. use the "exact" Gamma
        data without non-analytic correction.) This option takes priority over
        average_repeat_points.

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'qpts', 'weights', 'frequencies',
        'frequencies_unit'

        If read_eigenvectors is True, there will also be an
        'eigenvectors' key
    """
    with open(filename) as f:

        f.readline()  # Skip BEGIN header
        n_atoms = int(f.readline().split()[-1])
        n_branches = int(f.readline().split()[-1])
        n_qpts = int(f.readline().split()[-1])
        (cell_vectors, atom_r, atom_type, atom_mass) = _read_crystal_info(
            f, n_atoms)

        qpts = np.zeros((n_qpts, 3))
        weights = np.zeros(n_qpts)
        freqs = np.zeros((n_qpts, n_branches))
        if read_eigenvectors:
            eigenvecs = np.zeros((n_qpts, n_branches, n_atoms, 3),
                                 dtype='complex128')

        # Need to loop through file using while rather than number of
        # q-points as points are duplicated on LO-TO splitting
        previous_qpt_id = -1
        repeated_qpt_ids = defaultdict(set)
        loto_split_indices = set()

        for idx, frequency_block in enumerate(
                _read_frequency_blocks(f, n_branches)):
            qpt_id = frequency_block.qpt_id

            if prefer_non_loto and frequency_block.direction is not None:
                loto_split_indices.add(idx)

            if ((average_repeat_points or prefer_non_loto)
                and (qpt_id == previous_qpt_id)):
                repeated_qpt_ids[qpt_id].update({idx - 1, idx})
            previous_qpt_id = qpt_id

            [f.readline() for x in range(2)]  # Skip 2 label lines
            evec_lines = np.array(
                [f.readline() for x in range(n_atoms*n_branches)])
            if read_eigenvectors:
                evec_lines = np.array([x.split()[2:] for x in evec_lines],
                                      dtype=np.float64)
                lines_i = np.column_stack(
                    [evec_lines[:, 0] + evec_lines[:, 1]*1j,
                     evec_lines[:, 2] + evec_lines[:, 3]*1j,
                     evec_lines[:, 4] + evec_lines[:, 5]*1j])
                qeigenvec = np.zeros((n_branches, n_atoms, 3),
                                     dtype=np.complex128)
                for i in range(n_branches):
                    qeigenvec[i, :, :] = lines_i[i*n_atoms:(i+1)*n_atoms, :]
            # Sometimes there are more than n_qpts q-points in the file
            # due to LO-TO splitting
            if idx < len(qpts):
                qpts[idx] = frequency_block.qpt
                freqs[idx] = frequency_block.frequencies
                weights[idx] = frequency_block.weight
                if read_eigenvectors:
                    eigenvecs[idx] = qeigenvec
            else:
                qpts = np.concatenate((qpts, [frequency_block.qpt]))
                freqs = np.concatenate((freqs, [frequency_block.frequencies]))
                weights = np.concatenate((weights, [frequency_block.weight]))
                if read_eigenvectors:
                    eigenvecs = np.concatenate((eigenvecs, [qeigenvec]))

    # Multiple qpts with same CASTEP q-pt index: correct weights
    for indices in repeated_qpt_ids.values():
        if (prefer_non_loto
            and (intersection := indices & loto_split_indices)
            and len(indices) > len(intersection)):
            # Repeated q-point has both split and un-split variations;
            # set weights of split points to zero
            if len(indices) - len(intersection) == 1:
                zero_indices = np.asarray(list(intersection), dtype = int)
                weights[zero_indices] = 0
            else:
                msg = (
                    'Found multiple non-split blocks for q-point {qpt_id},'
                    ' cannot determine which to use.'
                )
                raise ValueError(msg)

        elif average_repeat_points:
            np_indices = np.asarray(list(indices), dtype=int)
            weights[np_indices] /= np_indices.size

    data_dict: dict[str, Any] = {}
    data_dict['crystal'] = {}
    cry_dict = data_dict['crystal']
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = (cell_vectors*ureg('angstrom').to(
        cell_vectors_unit)).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    cry_dict['atom_r'] = atom_r
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*(ureg('amu')).to(
        atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit
    data_dict['qpts'] = qpts
    data_dict['weights'] = weights
    data_dict['frequencies'] = ((freqs*(1/ureg.cm)).to(
        frequencies_unit)).magnitude
    data_dict['frequencies_unit'] = frequencies_unit
    if read_eigenvectors:
        data_dict['eigenvectors'] = eigenvecs

    return data_dict


def _read_crystal_info(
        f: TextIO, n_atoms: int,
        ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Reads the header crystal information from a CASTEP text file, from
    'Unit cell vectors' to 'END header'

    Parameters
    ----------
    f
        File object in read mode for the .phonon file containing the
        data
    n_atoms
        Number of atoms in the cell

    Returns
    -------
    cell_vectors
        Shape (3, 3) float ndarray. The unit cell vectors in Angstroms
    atom_r
        Shape (n_atoms, 3) float ndarray. The fractional position of
        each atom within the unit cell
    atom_type
        Shape (n_atoms,) string ndarray. The chemical symbols of each
        atom in the unit cell
    atom_mass
        Shape (n_atoms,) float ndarray. The mass of each atom in the
        unit cell in atomic mass units
    """
    while 'Unit cell vectors' not in f.readline():
        pass
    cell_vectors = np.array([[float(x) for x in f.readline().split()[0:3]]
                             for i in range(3)])
    f.readline()  # Skip fractional co-ordinates label
    atom_info = np.array([f.readline().split() for i in range(n_atoms)])
    atom_r = np.array([[float(x) for x in y[1:4]] for y in atom_info])
    atom_type = np.array([x[4] for x in atom_info])
    atom_mass = np.array([float(x[5]) for x in atom_info])
    f.readline()  # Skip END header line

    return cell_vectors, atom_r, atom_type, atom_mass


_qpt_index_pattern = re.compile(r'q-pt=\s*(\d+)')
_qpt_float_pattern = re.compile(r'-?\d+\.\d+')


class _FrequencyBlock(NamedTuple):
    qpt_id: int
    qpt: np.ndarray
    direction: Optional[np.ndarray]
    weight: float
    frequencies: np.ndarray
    extra: Optional[np.ndarray]


def _read_frequency_block(
    f: TextIO,
    n_branches: int,
    extra_columns: Optional[list] = None,
    terminator: str = '',
        ) -> _FrequencyBlock | None:
    """
    For a single q-point reads the q-point, weight, frequencies
    and optionally any extra columns

    Parameters
    ----------
    f
        File object in read mode for the file containing the data
    n_branches
        Expected number of frequencies (i.e. phonon branches)
    extra_columns
        The index(es) of extra columns to read after the frequencies
        column. e.g. to read the first column after frequencies
        extra_columns = [0]. To read the first and second columns after
        frequencies extra_columns = [0, 1]. For reference, in .phonon
        IR intensities = 0, Raman intensities = 1, and in .phonon_dos
        mode gradients = 0
    terminator
        If this line is reached, return None without raising an error

    Returns
    -------
    NamedTuple with attributes:

    qpt_id
        CASTEP-assigned index of the q-point
    qpt
        Shape (3,) float ndarray. The q-point in reciprocal fractional
        coordinates
    direction
        If q-point is a Gamma-point with non-analytic LO-TO splitting
        correction, this is the direction vector of that data. Otherwise, None.
    weight
        The weight of this q-point
    frequencies
        Shape (n_modes,) float ndarray. The phonon frequencies for this
        q-point in 1/cm
    extra
        None if extra_columns is None, otherwise a shape
        (len(extra_columns), n_modes) float ndarray
    """
    qpt_line = f.readline()
    if qpt_line == terminator:
        return None

    i_qpt = int(re.findall(_qpt_index_pattern, qpt_line)[0])
    floats = re.findall(_qpt_float_pattern, qpt_line)
    qpt = np.array([float(x) for x in floats[:3]])
    qweight = float(floats[3])

    direction: Optional[np.ndarray] = None
    if len(floats) >= 6:
        direction = floats[4:7]

    freq_lines = [f.readline().split()
                  for i in range(n_branches)]
    freq_col = 1
    qfreq = np.array([float(line[freq_col]) for line in freq_lines])
    extra: Optional[np.ndarray] = None
    if extra_columns is not None:
        if len(qpt_line.split()) > 6:
            # Indicates a split gamma point in .phonon, so there is an
            # additional column
            extra_columns = [x + 1 for x in extra_columns]
        extra = np.zeros((len(extra_columns), n_branches))
        for i, col in enumerate(extra_columns):
            extra[i] = np.array(
                [float(line[freq_col + col + 1]) for line in freq_lines])
    return _FrequencyBlock(i_qpt, qpt, direction, qweight, qfreq, extra)


def _read_frequency_blocks(*args, **kwargs) -> Iterator[_FrequencyBlock]:
    """Iterate over frequency blocks"""
    while True:
        frequency_block = _read_frequency_block(*args, **kwargs)
        if frequency_block is None:
            break
        yield frequency_block


def _read_castep_version(f: BinaryIO) -> Version:
    """Read next line and return CASTEP version; raise error if incompatible"""
    castep_version = Version(_read_entry(f).decode().strip())
    if castep_version < Version('17.1'):
        msg = dedent_and_fill("""
            Old castep file detected: Euphonic only supports post-Castep 17.1
            files.  Please rerun the calculation with a newer version of Castep
            with the original .cell file and a .castep file with a single line
            with the "continuation: <old.castep_bin>" keyword and use the new
            output .castep_bin file in Euphonic.""")
        raise ValueError(msg)
    return castep_version


def read_interpolation_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        force_constants_unit: str = 'hartree/bohr**2',
        born_unit: str = 'e',
        dielectric_unit: str = '(e**2)/(bohr*hartree)') -> dict[str, Any]:
    """
    Reads data from a .castep_bin or .check file and returns it in a
    dictionary

    Parameters
    ----------
    filename
        The path and name of the file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    force_constants_unit
        The unit to return the force constants in
    born_unit
        The unit to return the Born charges in
    dielectric_unit
        The unit to return the dielectric permittivity tensor in

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'force_constants', 'force_constants_unit',
        'sc_matrix' and 'cell_origins'. Also contains 'born',
        'born_unit', 'dielectric' and 'dielectric_unit' if they are
        present in the .castep_bin or .check file.
    """
    with open(filename, 'rb') as f:
        int_type = '>i4'
        float_type = '>f8'
        first_cell_read = True
        while (header := _read_entry(f).strip()) != b'END':
            if header == b'BEGIN_PARAMETERS_DUMP':
                castep_version: Version = _read_castep_version(f)

            if header == b'BEGIN_UNIT_CELL':
                # CASTEP writes the cell twice: the first is the
                # geometry optimised cell, the second is the original
                # cell. We only want the geometry optimised cell.
                if first_cell_read:
                    (n_atoms, cell_vectors, atom_r, atom_mass,
                     atom_type) = _read_cell(f, int_type, float_type)
                    first_cell_read = False
            elif header == b'FORCE_CON':
                sc_matrix = np.transpose(np.reshape(
                    _read_entry(f, int_type), (3, 3)))
                n_cells_in_sc = int(np.rint(np.absolute(
                    np.linalg.det(sc_matrix))))
                # Transpose and reshape fc so it is indexed [nc, i, j]
                force_constants = np.ascontiguousarray(np.transpose(
                    np.reshape(_read_entry(f, float_type),
                               (n_cells_in_sc, 3*n_atoms, 3*n_atoms)),
                    axes=[0, 2, 1]))

                cell_origins = np.reshape(_read_entry(f, int_type),
                                          (n_cells_in_sc, 3))

                _ = _read_entry(f, int_type)  # FC row not used
            elif header == b'BORN_CHGS':
                born = np.reshape(
                    _read_entry(f, float_type), (n_atoms, 3, 3))

                if castep_version > Version('25.1'):
                    # Extra field marks if born charges were read from BORN
                    _ = _read_entry(f)

            elif header == b'DIELECTRIC':
                entry = _read_entry(f, float_type)
                dielectric = np.transpose(np.reshape(
                    entry, (3, 3)))

                if castep_version > Version('25.1'):
                    # Extra field marks if dielectric tensor was read from BORN
                    _ = _read_entry(f)

    data_dict: dict[str, Any] = {}
    cry_dict = data_dict['crystal'] = {}
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = cell_vectors*ureg(
        'bohr').to(cell_vectors_unit).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    # Normalise atom coordinates
    cry_dict['atom_r'] = atom_r - np.floor(atom_r)
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*ureg(
        'electron_mass').to(atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit

    # Set entries relating to 'FORCE_CON' block
    try:
        data_dict['force_constants'] = force_constants*ureg(
            'hartree/bohr**2').to(force_constants_unit).magnitude
        data_dict['force_constants_unit'] = force_constants_unit
        data_dict['sc_matrix'] = sc_matrix
        data_dict['cell_origins'] = cell_origins
    except NameError:
        msg = dedent_and_fill(f"""
            Force constants matrix could not be found in {filename}.


            Ensure PHONON_WRITE_FORCE_CONSTANTS: true and a PHONON_FINE_METHOD
            has been chosen when running CASTEP
            """)
        raise RuntimeError(msg) from None

    # Set entries relating to dipoles
    try:
        data_dict['born'] = born*ureg('e').to(born_unit).magnitude
        data_dict['born_unit'] = born_unit
        data_dict['dielectric'] = dielectric*ureg(
            'e**2/(hartree*bohr)').to(dielectric_unit).magnitude
        data_dict['dielectric_unit'] = dielectric_unit

    except UnboundLocalError:
        pass

    return data_dict


def _read_cell(file_obj: BinaryIO, int_type: str, float_type: str,
               ) -> tuple[int, np.ndarray, np.ndarray,
                          np.ndarray, np.ndarray]:
    """
    Read cell data from a .castep_bin or .check file

    Parameters
    ----------
    file_obj
        File object in read mode for the .castep_bin or .check file
    int_type
        Python struct format string describing the size and endian-ness
        of ints in the file
    float_type
        Python struct format string describing the size and endian-ness
        of floats in the file

    Returns
    -------
    n_atoms
        Number of atoms in the unit cell
    cell_vectors
        Shape (3, 3) float ndarray. The unit cell vectors in bohr
    atom_r
        Shape (n_atoms, 3) float ndarray. The fractional position of
        each atom within the unit cell
    atom_mass
        Shape (n_atoms,) float ndarray. The mass of each atom in the
        unit cell in units of electron mass
    atom_type
        Shape (n_atoms,) string ndarray. The chemical symbols of each
        atom in the unit cell
    """
    while (header := _read_entry(file_obj).strip()) != b'END_UNIT_CELL':
        if header == b'CELL%NUM_IONS':
            n_atoms = _read_entry(file_obj, int_type)
        elif header == b'CELL%REAL_LATTICE':
            cell_vectors = np.transpose(np.reshape(
                _read_entry(file_obj, float_type), (3, 3)))
        elif header == b'CELL%NUM_SPECIES':
            n_species = _read_entry(file_obj, int_type)
        elif header == b'CELL%NUM_IONS_IN_SPECIES':
            n_atoms_in_species = _read_entry(file_obj, int_type)
            if n_species == 1:
                n_atoms_in_species = np.array([n_atoms_in_species])
        elif header == b'CELL%IONIC_POSITIONS':
            max_atoms_in_species = max(n_atoms_in_species)
            atom_r_tmp = np.reshape(_read_entry(file_obj, float_type),
                                    (n_species, max_atoms_in_species, 3))
        elif header == b'CELL%SPECIES_MASS':
            atom_mass_tmp = _read_entry(file_obj, float_type)
            if n_species == 1:
                atom_mass_tmp = np.array([atom_mass_tmp])
        elif header == b'CELL%SPECIES_SYMBOL':
            # Need to decode binary string for Python 3 compatibility
            if n_species == 1:
                atom_type_tmp = [_read_entry(file_obj, 'S8')
                                 .strip().decode('utf-8')]
            else:
                atom_type_tmp = [x.strip().decode('utf-8')
                                 for x in _read_entry(file_obj, 'S8')]
    # Get atom_r in correct form
    # CASTEP stores atom positions as 3D array (3,
    # max_atoms_in_species, n_species) so need to slice data to get
    # correct information
    atom_begin = np.insert(np.cumsum(n_atoms_in_species[:-1]), 0, 0)
    atom_end = np.cumsum(n_atoms_in_species)
    atom_r = np.zeros((n_atoms, 3))
    for i in range(n_species):
        atom_r[atom_begin[i]:atom_end[i], :] = atom_r_tmp[
            i, :n_atoms_in_species[i], :]
    # Get atom_type in correct form
    atom_type = np.array([])
    atom_mass = np.array([])
    for at in range(n_species):
        atom_type = np.append(
            atom_type,
            [atom_type_tmp[at] for i in range(n_atoms_in_species[at])])
        atom_mass = np.append(
            atom_mass,
            [atom_mass_tmp[at] for i in range(n_atoms_in_species[at])])

    return n_atoms, cell_vectors, atom_r, atom_mass, atom_type


def _read_entry(file_obj: BinaryIO, dtype: str = '',
                ) -> str | int | float | np.ndarray:
    """
    Read a record from a Fortran binary file, including the beginning
    and end record markers and return the data inbetween

    Parameters
    ----------
    f
        File object in read mode for the Fortran binary file
    dtype
        String determining what order and type to unpack the bytes as.
        See 'Format Strings' in Python struct documentation

    Returns
    -------
    data
        Data type returned depends on dtype specified. If dtype is not
        specified, return type is a string. If there is more than one
        element in the record, it is returned as an ndarray of floats or
        integers

    """
    def record_mark_read(file_obj):
        # Read 4 byte Fortran record marker
        rawdata = file_obj.read(4)
        if rawdata == b'':
            msg = 'Problem reading binary file: unexpected EOF reached'
            raise EOFError(msg)
        return struct.unpack('>i', rawdata)[0]

    begin = record_mark_read(file_obj)
    if dtype:
        n_bytes = int(dtype[-1])
        n_elems = int(begin/n_bytes)
        if n_elems > 1:
            data = np.fromfile(file_obj, dtype=dtype, count=n_elems)
            if 'i' in dtype:
                data = data.astype(np.int32)
            elif 'f' in dtype:
                data = data.astype(np.float64)
        elif 'i' in dtype:
            data = struct.unpack('>i', file_obj.read(begin))[0]
        elif 'f' in dtype:
            data = struct.unpack('>d', file_obj.read(begin))[0]
        else:
            data = file_obj.read(begin)
    else:
        data = file_obj.read(begin)
    end = record_mark_read(file_obj)
    if begin != end:
        msg = ('Problem reading binary file: beginning and end record markers '
               'do not match')
        raise OSError(msg)

    return data



#! /usr/bin/env python
"""Script to compute version from git tags.

Provides 3 means of updating:

- python version.py

  Print computed version to stdout

- python version.py --dump

  Update version in `euphonic/version.py`

- python version.py --dist

  Update versions in both `euphonic/version.py` and in meson sdist build directory.
"""

import os
from pathlib import Path
import subprocess
import sys

if new_dir := os.getenv('MESON_SOURCE_ROOT'):
    os.chdir(new_dir)

gits = ['git']
if sys.platform == 'win32':
    gits += ['git.cmd', 'git.exe']

match sys.argv:
    case [_, '--dist']:
        COMMAND = 'dist'
    case [_, '--dump']:
        COMMAND = 'dump'
    case [_]:
        COMMAND = 'print'

version_file = Path(__file__).parent.parent / 'euphonic' / 'version.py'

for gitcmd in gits:
    try:
        print(f'Trying {gitcmd} ...', file=sys.stderr)
        proc = subprocess.run([gitcmd, 'describe', '--tags', '--dirty'],  # noqa: S603
                              capture_output=True, check=True, text=True,
                              )
    except FileNotFoundError:
        print(f'Tried {gitcmd}, File Not Found', file=sys.stderr)
        continue

    except (subprocess.CalledProcessError) as err:
        print(f'Tried {gitcmd}, returned: {err}', file=sys.stderr)
        print(f"Stdout: '{err.stdout.strip()}'", file=sys.stderr)
        print(f"Stderr: '{err.stderr.strip()}'", file=sys.stderr)
        continue

    version, *dirty = proc.stdout.strip().split('-')
    if dirty:
        version += f"+{dirty[0]}.{dirty[1]}{'.dirty' if len(dirty) > 2 and COMMAND != 'dump' else ''}"
    break

else:  # Can't use git
    print('All git implementations failed, reading version file', file=sys.stderr)
    version = version_file.read_text().split('=')[1].strip('"\n ')

match COMMAND:
    case 'dist':
        version_file.write_text(f'__version__ = "{version}"', encoding='utf-8')
        dist_path = os.getenv('MESON_DIST_ROOT')
        dist_version_file = Path(dist_path) / 'euphonic' / 'version.py'
        dist_version_file.parent.mkdir(parents=True, exist_ok=True)
        dist_version_file.write_text(f'__version__ = "{version}"', encoding='utf-8')

    case 'dump':
        version_file.write_text(f'__version__ = "{version}"', encoding='utf-8')

    case 'print':
        print(version)



"""Matplotlib stylesheets for plot styling"""
from importlib.resources import files

base_style = files(__package__) / 'base.mplstyle'
intensity_widget_style = files(__package__) / 'intensity_widget.mplstyle'



"""A consistent Pint unit registry for Euphonic Quantity objects.

They can also be imported directly 'from euphonic import Quantity, ureg';
this module exists to avoid import loops between the core modules of Euphonic.

This registry has the 'spectroscopy' context enabled by default to facilitate
conversion between energy and wavenumber. During tricky conversions between
reciprocals of energy and wavenumber, try enabling the
'reciprocal_spectroscopy' ('rs') context.

>>> y = 1. * ureg("1/meV")
>>> y.to("1/ (1/cm)", "rs")
<Quantity(0.123984198, 'centimeter')>

"""

from importlib.resources import files

from pint import UnitRegistry

from . import data

ureg = UnitRegistry()

ureg.load_definitions(files(data) / 'reciprocal_spectroscopy_definitions.txt')
ureg.enable_contexts('spectroscopy')
Quantity = ureg.Quantity

__all__ = ['Quantity', 'ureg']



"""Spectrum Collection classes"""
# pylint: disable=no-member

from abc import ABC, abstractmethod
import collections
from collections.abc import Callable, Generator, Sequence
import copy
from functools import partial, reduce
from itertools import product, repeat
import json
from numbers import Integral
from operator import contains
from typing import (
    Any,
    Literal,
    Optional,
    TypeVar,
    overload,
)

import numpy as np
from pint import Quantity
from toolz.dicttoolz import keyfilter, valmap
from toolz.functoolz import complement
from toolz.itertoolz import groupby, pluck
from typing_extensions import Self

from euphonic.broadening import ErrorFit, KernelShape
from euphonic.io import _obj_to_dict, _process_dict
from euphonic.readers.castep import read_phonon_dos_data
from euphonic.ureg import ureg
from euphonic.util import dedent_and_fill
from euphonic.validate import _check_constructor_inputs
from euphonic.version import __version__

from .base import (
    CallableQuantity,
    Spectrum,
    Spectrum1D,
    Spectrum2D,
    XTickLabels,
)
from .base import OneSpectrumMetadata as OneLineData

LineData = Sequence[OneLineData]
Metadata = dict[str, str | int | LineData]


class SpectrumCollectionMixin(ABC):
    """Help a collection of spectra work with "line_data" metadata file

    This is a Mixin to be inherited by Spectrum collection classes

    To avoid redundancy, spectrum collections store metadata in the form

    {"key1": value1, "key2", value2, "line_data": [{"key3": value3, ...},
                                                   {"key4": value4, ...}...]}

    - It is not guaranteed that all "lines" carry the same keys
    - No key should appear at both top-level and in line-data; any key-value
      pair at top level is assumed to apply to all lines
    - "lines" can actually correspond to N-D spectra, the notation was devised
      for multi-line plots of Spectrum1DCollection and then applied to other
      purposes.

    The _spectrum_axis class attribute determines which axis property contains
    the spectral data, and should be set by subclasses (i.e. to "y" or "z" for
    1D or 2D).
    """

    # Subclasses must define which axis contains the spectral data for
    # purposes of splitting, indexing, etc.
    # Python doesn't support abstract class attributes so we define a default
    # value, ensuring _something_ was set.
    _bin_axes = ('x',)
    _spectrum_axis = 'y'
    _item_type = Spectrum1D

    # Define some private methods which wrap this information into useful forms
    @classmethod
    def _spectrum_data_name(cls) -> str:
        return f'{cls._spectrum_axis}_data'

    @classmethod
    def _raw_spectrum_data_name(cls) -> str:
        return f'_{cls._spectrum_axis}_data'

    def _get_spectrum_data(self) -> Quantity:
        return getattr(self, self._spectrum_data_name())

    def _get_raw_spectrum_data(self) -> np.ndarray:
        return getattr(self, self._raw_spectrum_data_name())

    def _set_spectrum_data(self, data: Quantity) -> None:
        setattr(self, self._spectrum_data_name(), data)

    def _set_raw_spectrum_data(self, data: np.ndarray) -> None:
        setattr(self, self._raw_spectrum_data_name(), data)

    def _get_spectrum_data_unit(self) -> str:
        return getattr(self, f'{self._spectrum_data_name()}_unit')

    def _get_internal_spectrum_data_unit(self) -> str:
        return getattr(self, f'_internal_{self._spectrum_data_name()}_unit')

    @property
    def _bin_axis_names(self) -> list[str]:
        return [f'{axis}_data' for axis in self._bin_axes]

    @classmethod
    def _get_item_data(cls, item: Spectrum) -> Quantity:
        return getattr(item, f'{cls._spectrum_axis}_data')

    @classmethod
    def _get_item_raw_data(cls, item: Spectrum) -> np.ndarray:
        return getattr(item, f'_{cls._spectrum_axis}_data')

    @classmethod
    def _get_item_data_unit(cls, item: Spectrum) -> str:
        return getattr(item, f'{cls._spectrum_axis}_data_unit')

    def sum(self) -> Spectrum:
        """
        Sum collection to a single spectrum

        Returns
        -------
        summed_spectrum
            A single combined spectrum from all items in collection. Any
            metadata in 'line_data' not common across all spectra will be
            discarded
        """
        metadata = copy.deepcopy(self.metadata)
        metadata.pop('line_data', None)
        metadata.update(self._tidy_metadata())
        summed_s_data = ureg.Quantity(
            np.sum(self._get_raw_spectrum_data(), axis=0),
            units=self._get_internal_spectrum_data_unit(),
        ).to(self._get_spectrum_data_unit())

        bin_kwargs = {axis_name: getattr(self, axis_name)
                      for axis_name in self._bin_axis_names}

        return self._item_type(
            **bin_kwargs,
            **{self._spectrum_data_name(): summed_s_data},
            x_tick_labels=copy.copy(self.x_tick_labels),
            metadata=metadata,
        )

    # Required methods
    @classmethod
    @abstractmethod
    def from_spectra(
            cls, spectra: Sequence[Spectrum], *, unsafe: bool = False,
    ) -> Self:
        """Construct spectrum collection from a sequence of components

        If 'unsafe', some consistency checks may be skipped to improve
        performance; this should only be used when e.g. combining data iterated
        by another Spectrum collection

        """

    # Mixin methods
    def __len__(self):
        return self._get_raw_spectrum_data().shape[0]

    @overload
    def __getitem__(self, item: int) -> Spectrum: ...

    @overload
    def __getitem__(self, item: slice) -> Self: ...

    @overload
    def __getitem__(self, item: Sequence[int] | np.ndarray) -> Self: ...

    def __getitem__(
            self, item: Integral | slice | Sequence[Integral] | np.ndarray,
    ):
        self._validate_item(item)

        if isinstance(item, Integral):
            spectrum = self._item_type.__new__(self._item_type)
        else:
            # Pylint miscounts arguments when we call this staticmethod
            spectrum = self.__new__(type(self))  # pylint: disable=E1120

        self._set_item_data(spectrum, item)

        spectrum.x_tick_labels = self.x_tick_labels
        spectrum.metadata = self._get_item_metadata(item)

        return spectrum

    def _set_item_data(
            self,
            spectrum: Spectrum,
            item: Integral | slice | Sequence[Integral] | np.ndarray,
    ) -> None:
        """Write axis and spectrum data from self to Spectrum

        This is intended to set attributes on a 'bare' Spectrum created with
        __new__() from the parent SpectrumCollection.
        """

        for axis in self._bin_axes:
            for prop in ('_{}_data', '_internal_{}_data_unit', '{}_data_unit'):
                name = prop.format(axis)
                setattr(spectrum, name, copy.copy(getattr(self, name)))

        setattr(spectrum, self._raw_spectrum_data_name(),
                self._get_raw_spectrum_data()[item, :].copy())

        setattr(spectrum, f'_internal_{self._spectrum_data_name()}_unit',
                self._get_internal_spectrum_data_unit())
        setattr(spectrum, f'{self._spectrum_data_name()}_unit',
                self._get_spectrum_data_unit())

    def _validate_item(
            self, item: Integral | slice | Sequence[Integral] | np.ndarray,
    ) -> None:
        """Raise Error if index has inappropriate typing/ranges

        Raises
        ------
        IndexError
            Slice is not compatible with size of collection

        TypeError
            item specification does not have acceptable type; e.g. a sequence
            of float or bool was provided when ints are needed.

        """
        if isinstance(item, Integral):
            return
        if isinstance(item, slice):
            if (item.stop is not None) and (item.stop >= len(self)):
                msg = f'index "{item.stop}" out of range'
                raise IndexError(msg)
            return

        if not all(isinstance(i, Integral) for i in item):
            msg = (
                f'Index "{item}" should be an integer, slice '
                f'or sequence of ints'
            )
            raise TypeError(msg)

    @overload
    def _get_item_metadata(self, item: Integral) -> OneLineData:
        """Get a single metadata item with no line_data"""

    @overload
    def _get_item_metadata(self, item: slice | Sequence[Integral] | np.ndarray,
                           ) -> Metadata:
        """Get a metadata collection (may include line_data)"""

    def _get_item_metadata(self, item):
        """Produce appropriate metadata for __getitem__"""
        metadata_lines = list(self.iter_metadata())

        if isinstance(item, Integral):
            return metadata_lines[item]
        if isinstance(item, slice):
            return self._combine_metadata(metadata_lines[item])
        # Item must be some kind of integer sequence
        return self._combine_metadata([metadata_lines[i] for i in item])

    def copy(self) -> Self:
        """Get an independent copy of spectrum"""
        return self._item_type.copy(self)

    def __add__(self, other: Self) -> Self:
        """
        Appends the y_data of 2 Spectrum1DCollection objects,
        creating a single Spectrum1DCollection that contains
        the spectra from both objects. The two objects must
        have equal x_data axes, and their y_data must
        have compatible units and the same number of y_data
        entries

        Any metadata key/value pairs that are common to both
        spectra are retained in the top level dictionary, any
        others are put in the individual 'line_data' entries
        """
        return type(self).from_spectra([*self, *other])

    def iter_metadata(self) -> Generator[OneLineData, None, None]:
        """Iterate over metadata dicts of individual spectra from collection"""
        common_metadata = {key: value for key, value in self.metadata.items()
                           if key != 'line_data'}

        line_data = self.metadata.get('line_data')
        if line_data is None:
            line_data = repeat({}, len(self._get_raw_spectrum_data()))

        for one_line_data in line_data:
            yield common_metadata | one_line_data

    def _select_indices(self, **select_key_values) -> list[int]:
        """Get indices of items that match metadata query

        The target key-value pairs are a subset of the matching data, e.g.

        self._select_indices(species="Na", weight="coherent")

        will match metadata rows

        {"species": "Na", "weight": "coherent"}

        and

        {"species": "Na", "weight": "coherent", "mass": "22.9898"}

        but not

        {"species": "Na"}

        or

        {"species": "K", "weight": "coherent"}
        """
        required_metadata = select_key_values.items()
        return [i for i, row in enumerate(self.iter_metadata())
                   if required_metadata <= row.items()]

    def select(
            self,
            **select_key_values: str | int | Sequence[str] | Sequence[int],
    ) -> Self:
        """
        Select spectra by their keys and values in metadata['line_data']

        Parameters
        ----------
        **select_key_values
            Key-value/values pairs in metadata['line_data'] describing
            which spectra to extract. For example, to select all spectra
            where metadata['line_data']['species'] = 'Na' or 'Cl' use
            spectrum.select(species=['Na', 'Cl']). To select 'Na' and
            'Cl' spectra where weighting is also coherent, use
            spectrum.select(species=['Na', 'Cl'], weighting='coherent')

        Returns
        -------
        selected_spectra
           A Spectrum1DCollection containing the selected spectra

        Raises
        ------
        ValueError
            If no matching spectra are found
        """
        # Convert all items to sequences of possibilities
        def ensure_sequence(value: int | str | Sequence[int | str],
                            ) -> Sequence[int | str]:
            return (value,) if isinstance(value, (int, str)) else value

        select_key_values = valmap(ensure_sequence, select_key_values)

        # Collect indices that match each combination of values
        selected_indices = []
        for value_combination in product(*select_key_values.values()):
            selection = dict(
                zip(select_key_values.keys(), value_combination, strict=True))
            selected_indices.extend(self._select_indices(**selection))

        if not selected_indices:
            msg = (
                f'No spectra found with matching metadata '
                f'for {select_key_values}'
            )
            raise ValueError(msg)

        return self[selected_indices]

    @staticmethod
    def _combine_metadata(all_metadata: LineData) -> Metadata:
        """
        From a sequence of metadata dictionaries, combines all common
        key/value pairs into the top level of a metadata dictionary,
        all unmatching key/value pairs are put into the 'line_data'
        key, which is a list of metadata dicts for each element in
        all_metadata
        """
        # This is for combining multiple separate spectrum metadata,
        # they shouldn't have line_data
        for metadata in all_metadata:
            assert 'line_data' not in metadata

        # Combine key-value pairs common to *all* metadata lines into new dict
        common_metadata = dict(
            reduce(set.intersection,
                   (set(metadata.items()) for metadata in all_metadata)))

        # Put all other per-spectrum metadata in line_data
        is_common = partial(contains, common_metadata)
        line_data = [keyfilter(complement(is_common), one_line_data)
                     for one_line_data in all_metadata]

        if any(line_data):
            return common_metadata | {'line_data': line_data}

        return common_metadata

    def _tidy_metadata(self) -> Metadata:
        """
        For a metadata dictionary, combines all common key/value
        pairs in 'line_data' and puts them in a top-level dictionary.
        """
        line_data = self.metadata.get('line_data', [{}] * len(self))
        combined_line_data = self._combine_metadata(line_data)
        combined_line_data.pop('line_data', None)
        return combined_line_data

    def _check_metadata(self) -> None:
        """Check self.metadata['line_data'] is consistent with collection size

        Raises
        ------
        ValueError
            Metadata contains 'line_data' of incorrect length

        """
        if 'line_data' in self.metadata:
            collection_size = len(self._get_raw_spectrum_data())
            n_lines = len(self.metadata['line_data'])

            if n_lines != collection_size:
                msg = dedent_and_fill(f"""
                    {self._spectrum_data_name()} contains {collection_size}
                    spectra, but metadata["line_data"] contains {n_lines}
                    entries
                """)
                raise ValueError(msg)

    def group_by(self, *line_data_keys: str) -> Self:
        """
        Group and sum elements of spectral data according to the values
        mapped to the specified keys in metadata['line_data']

        Parameters
        ----------
        line_data_keys
            The key(s) to group by. If only one line_data_key is
            supplied, if the value mapped to a key is the same for
            multiple spectra, they are placed in the same group and
            summed. If multiple line_data_keys are supplied, the values
            must be the same for all specified keys for them to be
            placed in the same group

        Returns
        -------
        grouped_spectrum
            A new Spectrum1DCollection with one line for each group. Any
            metadata in 'line_data' not common across all spectra in a
            group will be discarded
        """
        def get_key_items(enumerated_metadata: tuple[int, OneLineData],
                          ) -> tuple[str | int, ...]:
            """Get sort keys from an item of enumerated input to groupby

            e.g. with line_data_keys=("a", "b")

              (0, {"a": 4, "d": 5}) --> (4, None)
            """
            return tuple(enumerated_metadata[1].get(item, None)
                         for item in line_data_keys)

        # First element of each tuple is the index
        indices = partial(pluck, 0)

        groups = groupby(get_key_items, enumerate(self.iter_metadata()))

        return self.from_spectra([self[list(indices(group))].sum()
                                  for group in groups.values()])

    def to_dict(self) -> dict[str, Any]:
        """
        Convert to a dictionary consistent with from_dict()

        Returns
        -------
        dict
        """
        attrs = [*self._bin_axis_names,
                 self._spectrum_data_name(),
                 'x_tick_labels',
                 'metadata']

        return _obj_to_dict(self, attrs)

    @classmethod
    def from_dict(cls: Self, d: dict) -> Self:
        """Initialise a Spectrum Collection object from dict"""
        data_keys = [f'{dim}_data' for dim in cls._bin_axes]
        data_keys.append(cls._spectrum_data_name())

        d = _process_dict(d,
                          quantities=data_keys,
                          optional=['x_tick_labels', 'metadata'])

        data_args = [d[key] for key in data_keys]
        return cls(*data_args,
                   x_tick_labels=d['x_tick_labels'],
                   metadata=d['metadata'])

    @classmethod
    def _item_type_check(cls, spectrum) -> None:
        if not isinstance(spectrum, cls._item_type):
            msg = f'Item is not of type {cls._item_type.__name__}.'
            raise TypeError(msg)


class Spectrum1DCollection(SpectrumCollectionMixin,
                           Spectrum,
                           collections.abc.Sequence):
    """A collection of Spectrum1D with common x_data and x_tick_labels

    Intended for convenient storage of band structures, projected DOS
    etc. This object can be indexed or iterated to obtain individual
    Spectrum1D.

    Attributes
    ----------
    x_data
        Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The x_data
        points (if size == (n_x_data,)) or x_data bin edges (if size
        == (n_x_data + 1,))
    y_data
        Shape (n_entries, n_x_data) float Quantity. The plot data in y,
        in rows corresponding to separate 1D spectra
    x_tick_labels
        Sequence[Tuple[int, str]] or None. Special tick labels e.g. for
        high-symmetry points. The int refers to the index in x_data the
        label should be applied to
    metadata
        Dict[str, Union[int, str, LineData]] or None. Contains metadata
        about the spectra. Keys should be strings and values should be
        strings or integers.
        There are some functional keys:

          - 'line_data' : LineData
                          This is a Sequence[Dict[str, Union[int, str]],
                          it contains metadata for each spectrum in
                          the collection, and must be of length
                          n_entries
    """
    T = TypeVar('T', bound='Spectrum1DCollection')

    # Private attributes used by SpectrumCollectionMixin
    _spectrum_axis = 'y'
    _item_type = Spectrum1D

    def __init__(
            self, x_data: Quantity, y_data: Quantity,
            x_tick_labels: Optional[XTickLabels] = None,
            metadata: Optional[dict[str, str | int | LineData]] = None,
    ) -> None:
        """
        Parameters
        ----------
        x_data
            Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The
            x_data points (if size == (n_x_data,)) or x_data bin edges
            (if size == (n_x_data + 1,))
        y_data
            Shape (n_entries, n_x_data) float Quantity. The plot data
            in y, in rows corresponding to separate 1D spectra
        x_tick_labels
            Special tick labels e.g. for high-symmetry points. The int
            refers to the index in x_data the label should be applied to
        metadata
            Contains metadata about the spectra. Keys should be
            strings and values should be strings or integers.
            There are some functional keys:

              - 'line_data' : LineData
                              This is a Sequence[Dict[str, Union[int, str]],
                              it contains metadata for each spectrum
                              in the collection, and must be of length
                              n_entries
        """

        _check_constructor_inputs(
            [y_data, x_tick_labels, metadata],
            [Quantity, [list, type(None)], [dict, type(None)]],
            [(-1, -1), (), ()],
            ['y_data', 'x_tick_labels', 'metadata'])
        ny = len(y_data[0])
        _check_constructor_inputs(
            [x_data], [Quantity],
            [[(ny,), (ny+1,)]], ['x_data'])

        self._set_data(x_data, 'x')
        self._set_data(y_data, 'y')
        self.x_tick_labels = x_tick_labels

        self.metadata = metadata if metadata is not None else {}
        self._check_metadata()

    def _split_by_indices(self,
                          indices: Sequence[int] | np.ndarray,
                          ) -> list[Self]:
        """Split data along x-axis at given indices"""

        ranges = self._ranges_from_indices(indices)

        return [type(self)(self.x_data[x0:x1], self.y_data[:, x0:x1],
                           x_tick_labels=self._cut_x_ticks(self.x_tick_labels,
                                                           x0, x1),
                           metadata=self.metadata)
                for x0, x1 in ranges]

    @staticmethod
    def _from_spectra_data_check(spectrum, x_data, y_data_units, x_tick_labels,
                                 ) -> None:
        if (spectrum.y_data.units != y_data_units
                or spectrum.x_data.units != x_data.units):
            msg = 'Spectrum units in sequence are inconsistent'
            raise ValueError(msg)
        if not np.allclose(spectrum.x_data, x_data):
            msg = 'x_data in sequence are inconsistent'
            raise ValueError(msg)
        if spectrum.x_tick_labels != x_tick_labels:
            msg = 'x_tick_labels in sequence are inconsistent'
            raise ValueError(msg)

    @classmethod
    def from_spectra(
            cls: Self, spectra: Sequence[Spectrum1D], *, unsafe: bool = False,
    ) -> Self:
        """Combine Spectrum1D to produce a new collection

        x_bins and x_tick_labels must be consistent (in magnitude and units)
        across the input spectra. If 'unsafe', this will not be checked.

        """
        if len(spectra) < 1:
            msg = 'At least one spectrum is needed for collection'
            raise IndexError(msg)

        cls._item_type_check(spectra[0])
        x_data = spectra[0].x_data
        x_tick_labels = spectra[0].x_tick_labels
        y_data_length = len(spectra[0].y_data)
        y_data_magnitude = np.empty((len(spectra), y_data_length))
        y_data_magnitude[0, :] = spectra[0].y_data.magnitude
        y_data_units = spectra[0].y_data.units

        for i, spectrum in enumerate(spectra[1:]):
            if not unsafe:
                cls._item_type_check(spectrum)
                cls._from_spectra_data_check(
                    spectrum, x_data, y_data_units, x_tick_labels)

            y_data_magnitude[i + 1, :] = spectrum.y_data.magnitude

        metadata = cls._combine_metadata([spec.metadata for spec in spectra])
        y_data = ureg.Quantity(y_data_magnitude, y_data_units)
        return cls(x_data, y_data, x_tick_labels=x_tick_labels,
                   metadata=metadata)

    def to_text_file(self, filename: str,
                     fmt: Optional[str | Sequence[str]] = None) -> None:
        """
        Write to a text file. The header contains metadata and unit
        information, the first column is x_data and each subsequent
        column is a y_data spectrum. Note that text files written in
        this format cannot be read back in by Euphonic.

        Parameters
        ----------
        filename
            Name of the text file to write to
        fmt
            A format specifier or sequence of specifiers (one for each
            column), to be passed to numpy.savetxt
        """
        common_metadata = copy.deepcopy(self.metadata)
        line_data = common_metadata.pop('line_data',
                                        [{} for i in self.y_data])
        header = [f'Generated by Euphonic {__version__}',
                  f'x_data in ({self.x_data.units})',
                  f'y_data in ({self.y_data.units})',
                  f'Common metadata: {json.dumps(common_metadata)}',
                  r'Column 1: x_data']
        for i, line in enumerate(line_data):
            header += [f'Column {i + 2}: y_data[{i}] {json.dumps(line)}']
        out_data = np.hstack((self.get_bin_centres().magnitude[:, np.newaxis],
                              self.y_data.transpose().magnitude))
        kwargs = {'header': '\n'.join(header)}
        if fmt is not None:
            kwargs['fmt'] = fmt
        np.savetxt(filename, out_data, **kwargs)

    @classmethod
    def from_castep_phonon_dos(cls: type[T], filename: str) -> T:
        """
        Reads total DOS and per-element PDOS from a CASTEP
        .phonon_dos file

        Parameters
        ----------
        filename
            The path and name of the .phonon_dos file to read
        """
        data = read_phonon_dos_data(filename)
        n_spectra = len(data['dos'].keys())
        dos_size = len(next(iter(data['dos'].values())))
        y_data = np.zeros((n_spectra, dos_size))
        metadata = {'line_data': [{} for x in range(n_spectra)]}

        for i, (species, dos_data) in enumerate(data['dos'].items()):
            y_data[i] = dos_data
            if species != 'Total':
                metadata['line_data'][i]['species'] = species
            metadata['line_data'][i]['label'] = species
        return Spectrum1DCollection(
            ureg.Quantity(data['dos_bins'], units=data['dos_bins_unit']),
            ureg.Quantity(y_data, units=data['dos_unit']),
            metadata=metadata)

    @overload
    def broaden(self: T, x_width: Quantity,
                shape: KernelShape = 'gauss',
                method: Optional[Literal['convolve']] = None,
                ) -> T: ...

    @overload
    def broaden(self: T, x_width: CallableQuantity,
                shape: KernelShape = 'gauss',
                method: Optional[Literal['convolve']] = None,
                width_lower_limit: Optional[Quantity] = None,
                width_convention: Literal['fwhm', 'std'] = 'fwhm',
                width_interpolation_error: float = 0.01,
                width_fit: ErrorFit = 'cheby-log',
                ) -> T: ...

    def broaden(self: T,
                x_width,
                shape='gauss',
                method=None,
                width_lower_limit=None,
                width_convention='fwhm',
                width_interpolation_error=0.01,
                width_fit='cheby-log',
                ) -> T:
        """
        Individually broaden each line in y_data, returning a new
        Spectrum1DCollection

        Parameters
        ----------
        x_width
            Scalar float Quantity, giving broadening FWHM for all y data, or
            a function handle accepting Quantity consistent with x-axis as
            input and returning FWHM corresponding to input array. This would
            typically be an energy-dependent resolution function.
        shape
            One of {'gauss', 'lorentz'}. The broadening shape
        method
            Can be None or 'convolve'. Currently the only broadening
            method available is convolution with a broadening kernel
            but this may not produce correct results for unequal bin
            widths. To use convolution anyway, explicitly set
            method='convolve'
        width_lower_limit
            Set a lower bound to width obtained calling x_width function. By
            default, this is equal to bin width. To disable, set to -Inf.
        width_convention
            By default ('fwhm'), x_width is interpreted as full-width
            half-maximum. Set to 'std' to instead define standard deviation.
        width_interpolation_error
            When x_width is a callable function, variable-width broadening is
            implemented by an approximate kernel-interpolation scheme. This
            parameter determines the target error of the kernel approximations.
        width_fit
            Select parametrisation of kernel width spacing to
            width_interpolation_error. 'cheby-log' is recommended: for shape
            'gauss', 'cubic' is also available.

        Returns
        -------
        broadened_spectrum
            A new Spectrum1DCollection object with broadened y_data

        Raises
        ------
        ValueError
            If shape is not one of the allowed strings
        ValueError
            If method is None and bins are not of equal size
        """
        if isinstance(x_width, Quantity):
            y_broadened = np.zeros_like(self.y_data)
            x_centres = [self.get_bin_centres().magnitude]
            x_width_calc = [x_width.to(self.x_data_unit).magnitude]
            for i, yi in enumerate(self.y_data.magnitude):
                y_broadened[i] = self._broaden_data(
                    yi, x_centres, x_width_calc, shape=shape,
                    method=method, width_convention=width_convention)

            new_spectrum = self.copy()
            new_spectrum.y_data = ureg.Quantity(
                y_broadened, units=self.y_data_unit)
            return new_spectrum

        if isinstance(x_width, Callable):
            return type(self).from_spectra([
                spectrum.broaden(
                    x_width=x_width,
                    shape=shape,
                    method=method,
                    width_lower_limit=width_lower_limit,
                    width_convention=width_convention,
                    width_interpolation_error=width_interpolation_error,
                    width_fit=width_fit)
                for spectrum in self])

        msg = 'x_width must be a Quantity or Callable'
        raise TypeError(msg)

    @classmethod
    def from_dict(cls: Self, d: dict) -> Self:
        """
        Convert a dictionary to a Spectrum Collection object

        Parameters
        ----------
        d : dict
            A dictionary with the following keys/values:

            - 'x_data': (n_x_data,) or (n_x_data + 1,) float ndarray
            - 'x_data_unit': str
            - 'y_data': (n_x_data,) float ndarray
            - 'y_data_unit': str

            There are also the following optional keys:

            - 'x_tick_labels': list of (int, string) tuples
            - 'metadata': dict

        Returns
        -------
        spectrum_collection
        """
        return super().from_dict(d)


class Spectrum2DCollection(SpectrumCollectionMixin,
                           Spectrum,
                           collections.abc.Sequence):
    """A collection of Spectrum2D with common x_data, y_data and x_tick_labels

    Intended for convenient storage of contributions to spectral maps such as
    S(Q,w). This object can be indexed or iterated to obtain individual
    Spectrum2D.

    Attributes
    ----------
    x_data
        Shape (n_x_data,) or (n_x_data + 1,) float Quantity. The x_data
        points (if size == (n_x_data,)) or x_data bin edges (if size
        == (n_x_data + 1,))
    y_data
        Shape (n_y_data,) or (n_y_data + 1,) float Quantity. The y_data
        points (if size == (n_y_data,)) or y_data bin edges (if size
        == (n_y_data + 1,))
    z_data
        Shape (n_entries, n_x_data, n_y_data) float Quantity. The spectral data
        in x and y, indexed over components
    x_tick_labels
        Sequence[Tuple[int, str]] or None. Special tick labels e.g. for
        high-symmetry points. The int refers to the index in x_data the
        label should be applied to
    metadata
        Dict[str, Union[int, str, LineData]] or None. Contains metadata
        about the spectra. Keys should be strings and values should be
        strings or integers.
        There are some functional keys:

          - 'line_data' : LineData
                          This is a Sequence[Dict[str, Union[int, str]],
                          it contains metadata for each spectrum in
                          the collection, and must be of length
                          n_entries
    """

    # Private attributes used by SpectrumCollectionMixin
    _bin_axes = ('x', 'y')
    _spectrum_axis = 'z'
    _item_type = Spectrum2D

    def __init__(
            self, x_data: Quantity, y_data: Quantity, z_data: Quantity,
            x_tick_labels: Optional[XTickLabels] = None,
            metadata: Optional[Metadata] = None,
    ) -> None:
        _check_constructor_inputs(
            [z_data, x_tick_labels, metadata],
            [Quantity, [list, type(None)], [dict, type(None)]],
            [(-1, -1, -1), (), ()],
            ['z_data', 'x_tick_labels', 'metadata'])
        # First axis corresponds to spectra in collection
        _, nx, ny = z_data.shape
        _check_constructor_inputs(
            [x_data, y_data],
            [Quantity, Quantity],
            [[(nx,), (nx + 1,)], [(ny,), (ny + 1,)]],
            ['x_data', 'y_data'])

        self._set_data(x_data, 'x')
        self._set_data(y_data, 'y')
        self.x_tick_labels = x_tick_labels
        self._set_data(z_data, 'z')

        self.metadata = metadata if metadata is not None else {}
        self._check_metadata()

    def _split_by_indices(self, indices: Sequence[int] | np.ndarray,
                          ) -> list[Self]:
        """Split data along x axis at given indices"""
        ranges = self._ranges_from_indices(indices)
        return [type(self)(self.x_data[x0:x1],
                           self.y_data,
                           self.z_data[:, x0:x1, :],
                           x_tick_labels=self._cut_x_ticks(
                               self.x_tick_labels, x0, x1),
                           metadata=self.metadata)
                for x0, x1 in ranges]

    @property
    def z_data(self) -> Quantity:
        """intensity data"""
        return ureg.Quantity(
            self._z_data, self._internal_z_data_unit,
        ).to(self.z_data_unit, 'reciprocal_spectroscopy')

    @z_data.setter
    def z_data(self, value: Quantity) -> None:
        self.z_data_unit = str(value.units)
        self._z_data = value.to(self._internal_z_data_unit).magnitude

    @staticmethod
    def _from_spectra_data_check(
            spectrum_0_data_units, spectrum_i_data_units,
            x_tick_labels, spectrum_i_x_tick_labels,
    ) -> None:
        """Check spectrum data units and x_tick_labels are consistent"""
        if spectrum_0_data_units != spectrum_i_data_units:
            msg = 'Spectrum units in sequence are inconsistent'
            raise ValueError(msg)
        if x_tick_labels != spectrum_i_x_tick_labels:
            msg = 'x_tick_labels in sequence are inconsistent'
            raise ValueError(msg)

    @staticmethod
    def _from_spectra_bins_check(ref_bins_data: dict[str, Quantity],
                                 spectrum: Spectrum2D) -> None:
        """Check bin values and units match between spectra in new collection

        Args:
            ref_bins_data: ref axis names and values in format
                ``{"x_data": Quantity(...), }``
            spectrum: Item from sequence to compare with ref
        """
        for key, ref_bins in ref_bins_data.items():
            item_bins = getattr(spectrum, key)
            if not np.allclose(item_bins.magnitude, ref_bins.magnitude):
                msg = 'Bins in sequence are inconsistent'
                raise ValueError(msg)
            if item_bins.units != ref_bins.units:
                msg = 'Bin units in sequence are inconsistent'
                raise ValueError(msg)

    @classmethod
    def from_spectra(
            cls, spectra: Sequence[Spectrum2D], *, unsafe: bool = False,
    ) -> Self:
        """Combine Spectrum2D to produce a new collection

        bins and x_tick_labels must be consistent (in magnitude and units)
        across the input spectra. If 'unsafe', this will not be checked.

        """

        if len(spectra) < 1:
            msg = 'At least one spectrum is needed for collection'
            raise IndexError(msg)

        cls._item_type_check(spectra[0])
        bins_data = {
            f'{ax}_data': getattr(spectra[0], f'{ax}_data')
            for ax in cls._bin_axes
        }
        x_tick_labels = spectra[0].x_tick_labels

        spectrum_0_data = cls._get_item_data(spectra[0])
        spectrum_data_shape = spectrum_0_data.shape
        spectrum_data_magnitude = np.empty(
            (len(spectra), *spectrum_data_shape))
        spectrum_data_magnitude[0, :, :] = spectrum_0_data.magnitude
        spectrum_data_units = spectrum_0_data.units

        for i, spectrum in enumerate(spectra[1:], start=1):
            spectrum_i_raw_data = cls._get_item_raw_data(spectrum)
            spectrum_i_data_units = cls._get_item_data_unit(spectrum)

            if not unsafe:
                cls._item_type_check(spectrum)
                cls._from_spectra_data_check(
                    spectrum_data_units, spectrum_i_data_units,
                    x_tick_labels, spectrum.x_tick_labels)
                cls._from_spectra_bins_check(bins_data, spectrum)

            spectrum_data_magnitude[i, :, :] = spectrum_i_raw_data

        metadata = cls._combine_metadata([spec.metadata for spec in spectra])
        spectrum_data = ureg.Quantity(spectrum_data_magnitude,
                                      spectrum_data_units)
        return cls(**bins_data,
                   **{f'{cls._spectrum_axis}_data': spectrum_data},
                   x_tick_labels=x_tick_labels,
                   metadata=metadata)

    def get_bin_edges(
            self,
            bin_ax: Literal['x', 'y'] = 'x',
            *,
            restrict_range: bool = True,
    ) -> Quantity:
        """
        Get bin edges for the axis specified by bin_ax. If the size of bin_ax
        is one element larger than z_data along that axis, bin_ax is assumed to
        contain bin edges. If they are the same size, bin_ax is assumed to
        contain bin centres and a conversion is made.

        In this case, bin edges are assumed to be halfway between bin centres.

        Parameters
        ----------
        bin_ax
            The axis to get the bin edges for, 'x' or 'y'

        restrict_range
            If True (default), the bin edges will not go outside the existing
            data bounds so the first and last bins may be half-size. This may
            be desirable for plotting.  Otherwise, the outer bin edges will
            extend from the initial data range.
        """
        enum = {'series': 0, 'x': 1, 'y': 2}
        bin_data = getattr(self, f'{bin_ax}_data')
        data_ax_len = self.z_data.shape[enum[bin_ax]]
        if self._is_bin_edge(data_ax_len, bin_data.shape[0]):
            return bin_data
        return self._bin_centres_to_edges(
            bin_data, restrict_range=restrict_range)

    def get_bin_centres(self, bin_ax: Literal['x', 'y'] = 'x') -> Quantity:
        """
        Get bin centres for the axis specified by bin_ax. If the size of
        bin_ax is the same size as z_data along that axis, bin_ax is
        assumed to contain bin centres, but if bin_ax is one element
        larger, bin_ax is assumed to contain bin centres and a conversion
        is made. In this conversion, the bin centres are assumed to be in
        the middle of each bin, which may not be an accurate assumption in
        the case of differently sized bins.

        Parameters
        ----------
        bin_ax
            The axis to get the bin centres for, 'x' or 'y'
        """
        enum = {'series': 0, 'x': 1, 'y': 2}
        bin_data = getattr(self, f'{bin_ax}_data')
        data_ax_len = self.z_data.shape[enum[bin_ax]]
        if self._is_bin_edge(data_ax_len, bin_data.shape[0]):
            return self._bin_edges_to_centres(bin_data)
        return bin_data



"""Export to JSON for phonon visualisation website:

- https://henriquemiranda.github.io/phononwebsite/index.html
"""

from collections import Counter
from itertools import pairwise
import json
from pathlib import Path
from typing import Any, TypedDict

import numpy as np

from euphonic.crystal import Crystal
from euphonic.qpoint_phonon_modes import QpointPhononModes
from euphonic.spectra import XTickLabels
from euphonic.util import _calc_abscissa, get_qpoint_labels

ComplexPair = tuple[float, float]


class PhononWebsiteData(TypedDict):
    """Data container for export to phonon visualisation website

    Specification: https://henriquemiranda.github.io/phononwebsite/index.html

    line_breaks are currently not implemented

    """
    name: str
    natoms: int
    lattice: list[list[float]]
    atom_types: list[str]
    atom_numbers: list[int]
    formula: str
    repetitions: list[int]
    atom_pos_car: list[list[float]]
    atom_pos_red: list[list[float]]
    highsym_qpts: list[tuple[int, str]]
    qpoints: list[list[float]]
    distances: list[float]  # Cumulative distance from first q-point
    eigenvalues: list[float]  # in cm-1
    vectors: list[list[list[tuple[ComplexPair, ComplexPair, ComplexPair]]]]
    line_breaks: list[tuple[int, int]]


def write_phonon_website_json(
    modes: QpointPhononModes,
    output_file: str | Path = 'phonons.json',
    name: str = 'Euphonic export',
    x_tick_labels: XTickLabels | None = None,
) -> None:

    """Dump to .json for use with phonon website visualiser

    Use with javascript application at
    https://henriquemiranda.github.io/phononwebsite

    Parameters
    ----------
    modes
        Phonon frequencies and eigenvectors along q-point path
    output_file
        Path to output file
    name
        Set "name" metadata, to be used as figure title
    x_tick_labels
        index and label for high symmetry labels (if known)

    """

    with open(output_file, 'w', encoding='utf-8') as fd:
        json.dump(_modes_to_phonon_website_dict(modes=modes,
                                                name=name,
                                                x_tick_labels=x_tick_labels),
                  fd)


def _crystal_website_data(crystal: Crystal) -> dict[str, Any]:
    elements = [
        '_', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na',
        'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V',
        'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se',
        'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh',
        'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba',
        'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho',
        'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt',
        'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac',
        'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm',
        'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg',
        'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']

    def get_z(symbol: str) -> int:
        try:
            return elements.index(symbol)
        except ValueError:  # Symbol not found
            return 0

    def symbols_to_formula(symbols: list[str]) -> str:
        symbol_counts = Counter(symbols)

        return ''.join(f'{symbol}{symbol_counts[symbol]}'
                       for symbol in sorted(symbol_counts))

    return {
        'natoms': len(crystal.atom_type),
        'lattice': crystal.cell_vectors.to('angstrom').magnitude.tolist(),
        'atom_types': crystal.atom_type.tolist(),
        'atom_numbers': list(map(get_z, crystal.atom_type)),
        'formula': symbols_to_formula(crystal.atom_type),
        'atom_pos_red': crystal.atom_r.tolist(),
        'atom_pos_car': (crystal.atom_r @ crystal.cell_vectors
                      ).to('angstrom').magnitude.tolist(),
    }


def _remove_breaks(distances: np.ndarray, btol: float = 10.) -> list[int]:
    """Collapse large breaks in cumulative-distance array

    These correspond to discontinuous regions of the x-axis: in euphonic
    plotting this is usually handled by splitting the spectrum and plotting
    to new axes, but Phonon Website does not handle this.

    Data is modified in-place

    A list of identified breakpoints is returned

    """
    diff = np.diff(distances)
    median = np.median(diff)
    breakpoints = np.where((diff / median) > btol)[0] + 1

    for line_break in reversed(breakpoints):
        distances[line_break:] -= diff[line_break - 1]

    return breakpoints.tolist()


def _find_duplicates(distances: np.ndarray) -> list[int]:
    """Identify breakpoints where a q-point is repeated in list

    Return the higher index of two points: this works more nicely with phonon
    website interpretation of line break points.
    """

    diff = np.diff(distances)
    duplicates = np.where(np.isclose(diff, 0.))[0] + 1

    return duplicates.tolist()


def _combine_neighbouring_labels(x_tick_labels: XTickLabels) -> XTickLabels:
    """Merge neighbouring labels in x_tick_label list

    If labels are the same, only keep one.

    If labels are different, join with |

    e.g.::

      [(1, "X"), (2, "X"), (4, "A"), (7, "Y"), (8, "Z")]

      -->

      [(2, "X"), (4, "A"), (8, "Y|Z")]

    """
    labels = dict(x_tick_labels)

    for (index, next_index) in pairwise(sorted(labels)):

        if index + 1 == next_index:  # Labels are neighbours

            if labels[index] != labels[next_index]:  # Combine differing labels
                labels[next_index] = f'{labels[index]}|{labels[next_index]}'

            del labels[index]  # Drop redundant label

    return sorted(labels.items())


def _modes_to_phonon_website_dict(
    modes: QpointPhononModes,
    name: str | None = None,
    repetitions: tuple[int, int, int] = (2, 2, 2),
    x_tick_labels: XTickLabels | None = None,
) -> PhononWebsiteData:
    """Convert QpointPhononModes to dict of phonon-website-friendly data"""

    qpts = modes.qpts
    eigenvectors = modes.eigenvectors

    abscissa = _calc_abscissa(modes.crystal.reciprocal_cell(), qpts)

    duplicates = _find_duplicates(abscissa)
    breakpoints = _remove_breaks(abscissa)

    breakpoints = sorted({0, *duplicates, *breakpoints, len(abscissa)})
    line_breaks = list(pairwise(breakpoints))

    if x_tick_labels is None:
        x_tick_labels = get_qpoint_labels(qpts,
                                          cell=modes.crystal.to_spglib_cell())

    x_tick_labels = [(int(key) + 1, str(value))
                     for key, value in x_tick_labels]
    x_tick_labels = _combine_neighbouring_labels(x_tick_labels)

    mass_weights = 1 / np.sqrt(modes.crystal.atom_mass)
    vectors = eigenvectors * mass_weights[None, None, :, None]
    # Convert complex numbers to a final axis over (real, imag)
    vectors = vectors.view(float).reshape(*eigenvectors.shape[:-1], 3, 2)

    return PhononWebsiteData(
        name=name,
        **_crystal_website_data(modes.crystal),
        highsym_qpts=x_tick_labels,
        distances=abscissa.magnitude.tolist(),
        qpoints=modes.qpts.tolist(),
        eigenvalues=modes.frequencies.to('1/cm').magnitude.tolist(),
        vectors=vectors.tolist(),
        repetitions=repetitions,
        line_breaks=line_breaks,
    )




from .base import (
    CallableQuantity,
    Spectrum,
    Spectrum1D,
    Spectrum2D,
    XTickLabels,
    apply_kinematic_constraints,
)
from .collections import (
    LineData,
    Metadata,
    OneLineData,
    Spectrum1DCollection,
    Spectrum2DCollection,
)

__all__ = [
    'CallableQuantity',
    'LineData',
    'Metadata',
    'OneLineData',
    'Spectrum',
    'Spectrum1D',
    'Spectrum1DCollection',
    'Spectrum2D',
    'Spectrum2DCollection',
    'XTickLabels',
    'apply_kinematic_constraints',
]



#! /usr/bin/env python
import argparse
import json
import os
import re

from packaging.version import Version
import requests
import yaml

from euphonic import __version__

POST_TIMEOUT = 5.  # 5s timeout; otherwise failed request can hang


def main():
    parser = get_parser()
    args = parser.parse_args()

    test = not args.notest
    if args.github:
        release_github(test)


def release_github(test=True):
    with open('CHANGELOG.rst', encoding='utf8') as f:
        changelog = f.read()
    with open('CITATION.cff', encoding='utf8') as f:
        citation = yaml.safe_load(f)

    euphonic_ver = __version__
    is_prerelease = Version(euphonic_ver).is_prerelease

    version_dict = {}
    if not is_prerelease:
        version_dict['CHANGELOG.rst'] = re.findall(r'\n`(v\d+\.\d+\.\S+)\s',
                                                   changelog)[0]
    version_dict['CITATION.cff'] = 'v' + citation['version']
    for ver_name, ver in version_dict.items():
        if euphonic_ver != ver:
            msg = (
                f'euphonic.__version__/{ver_name} version mismatch! '
                f'euphonic.__version__: {euphonic_ver} {ver_name}: {ver}'
            )
            raise ValueError(msg)

    if is_prerelease:
        body_re = r'`Unreleased.*?^-+\n(.*?)^`v'
    else:
        body_re = r'`v\d+\.\d+\.\S+.*?^-+\n(.*?)^`v'

    desc = re.search(body_re, changelog,
                     re.DOTALL | re.MULTILINE).groups()[0].strip()

    payload = {
        'tag_name': euphonic_ver,
        'target_commitish': 'master',
        'name': euphonic_ver,
        'body': desc,
        'draft': False,
        'prerelease': is_prerelease,
    }
    if test:
        print(payload)

        if 'GITHUB_TOKEN' in os.environ:
            print('Found GITHUB_TOKEN')
        else:
            print('No GITHUB_TOKEN set')
    else:
        response = requests.post(
            'https://api.github.com/repos/pace-neutrons/euphonic/releases',
            data=json.dumps(payload),
            headers={'Authorization': 'token ' + os.environ['GITHUB_TOKEN']},
            timeout=POST_TIMEOUT)
        print(response.text)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--github',
        action='store_true',
        help='Release on Github')
    parser.add_argument(
        '--notest',
        action='store_true',
        help='Actually send/upload')
    return parser


if __name__ == '__main__':
    main()
